#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import logging
import threading

from flask import Flask, render_template

logger = logging.getLogger(__name__)

app = Flask(__name__)
stop_names = {
    "BKK_F00965": "Vörösmarty tér",
    "BKK_F00963": "Deák Ferenc tér",
    "BKK_F00997": "Bajcsy-Zsilinszky út",
    "BKK_F01080": "Opera",
    "BKK_F01086": "Oktogon",
    "BKK_F01093": "Vörösmarty utca",
    "BKK_F01096": "Kodály körönd",
    "BKK_F01101": "Bajza utca",
    "BKK_F01103": "Hősök tere",
    "BKK_F02697": "Széchenyi fürdő",
    "BKK_F02888": "Mexikói út",
    "BKK_F00094": "Déli pályaudvar",
    "BKK_F02481": "Szél Kálmán tér",
    "BKK_F00063": "Batthyány tér",
    "BKK_F00959": "Kossuth Lajos tér",
    "BKK_F00961": "Deák Ferenc tér",
    "BKK_F01019": "Astoria",
    "BKK_F01292": "Blaha Lujza tér",
    "BKK_F01336": "Keleti pályaudvar",
    "BKK_F01325": "Puskás Ferenc Stadion",
    "BKK_F01743": "Pillangó utca",
    "BKK_F01749": "Örs vezér tere",
    "BKK_F00913": "Újpest-központ",
    "BKK_F00898": "Újpest-városkapu",
    "BKK_F02685": "Gyöngyösi utca",
    "BKK_F02683": "Forgách utca",
    "BKK_F02638": "Göncz Árpád városközpont",
    "BKK_F02681": "Dózsa György út",
    "BKK_F02614": "Lehel tér",
    "BKK_F00937": "Nyugati pályaudvar",
    "BKK_F00957": "Arany János utca",
    "BKK_F00955": "Deák Ferenc tér",
    "BKK_F00953": "Ferenciek tere",
    "BKK_F01290": "Kálvin tér",
    "BKK_F01189": "Corvin-negyed",
    "BKK_F01233": "Klinikák",
    "BKK_F01253": "Nagyvárad tér",
    "BKK_F01283": "Népliget",
    "BKK_F01494": "Ecseri út",
    "BKK_F01879": "Pöttyös utca",
    "BKK_F01542": "Határ út",
    "BKK_F01544": "Kőbánya-Kispest",
    "BKK_056215": "Kelenföldi pályaudvar",
    "BKK_056217": "Bikás park",
    "BKK_056219": "Újbuda-Központ",
    "BKK_056221": "Móricz Zsigmond körtér",
    "BKK_056223": "Szent Gellért tér",
    "BKK_056225": "Fővám tér",
    "BKK_056227": "Kálvin tér",
    "BKK_056229": "Rákóczi tér",
    "BKK_056231": "II. János Pál pápa tér",
    "BKK_056233": "Keleti pályaudvar",
    "BKK_09001187": "Batthyány tér",
    "BKK_09019190": "Margit híd",
    "BKK_09043192": "Szépvölgyi út",
    "BKK_09050194": "Tímár utca",
    "BKK_09068196": "Szentlélek tér",
    "BKK_09118198": "Filatorigát",
    "BKK_09084200": "Kaszásdűlű",
    "BKK_09100202": "Aquincum",
    "BKK_09159204": "Rómaifürdő",
    "BKK_19720236": "Közvágóhíd",
    "BKK_19726238": "Kén utca",
    "BKK_19729241": "Pesterzsébet felső",
    "BKK_09220224": "Boráros tér",
    "BKK_09221226": "Müpa",
    "BKK_09223229": "Szabadkikötő",
    "BKK_19795278": "Örs vezér tere",
    "BKK_19798281": "Rákosfalva",
    "BKK_F00964": "Vörösmarty tér",
    "BKK_F00962": "Deák Ferenc tér",
    "BKK_F00996": "Bajcsy-Zsilinszky út",
    "BKK_F01079": "Opera",
    "BKK_F01085": "Oktogon",
    "BKK_F01092": "Vörösmarty utca",
    "BKK_F01095": "Kodály körönd",
    "BKK_F01100": "Bajza utca",
    "BKK_F01102": "Hősök tere",
    "BKK_F02696": "Széchenyi fürdő",
    "BKK_F02887": "Mexikói út",
    "BKK_F00093": "Déli pályaudvar",
    "BKK_F02480": "Szél Kálmán tér",
    "BKK_F00062": "Batthyány tér",
    "BKK_F00958": "Kossuth Lajos tér",
    "BKK_F00960": "Deák Ferenc tér",
    "BKK_F01018": "Astoria",
    "BKK_F01291": "Blaha Lujza tér",
    "BKK_F01335": "Keleti pályaudvar",
    "BKK_F01324": "Puskás Ferenc Stadion",
    "BKK_F01742": "Pillangó utca",
    "BKK_F00912": "Újpest-központ",
    "BKK_F00897": "Újpest-városkapu",
    "BKK_F02684": "Gyöngyösi utca",
    "BKK_F02682": "Forgách utca",
    "BKK_F02637": "Göncz Árpád városközpont",
    "BKK_F02680": "Dózsa György út",
    "BKK_F02613": "Lehel tér",
    "BKK_F00936": "Nyugati pályaudvar",
    "BKK_F00956": "Arany János utca",
    "BKK_F00954": "Deák Ferenc tér",
    "BKK_F00952": "Ferenciek tere",
    "BKK_F01289": "Kálvin tér",
    "BKK_F01188": "Corvin-negyed",
    "BKK_F01232": "Klinikák",
    "BKK_F01252": "Nagyvárad tér",
    "BKK_F01282": "Népliget",
    "BKK_F01493": "Ecseri út",
    "BKK_F01878": "Pöttyös utca",
    "BKK_F01541": "Határ út",
    "BKK_F01543": "Kőbánya-Kispest",
    "BKK_056216": "Kelenföldi pályaudvar",
    "BKK_056218": "Bikás park",
    "BKK_056220": "Újbuda-Központ",
    "BKK_056222": "Móricz Zsigmond körtér",
    "BKK_056224": "Szent Gellért tér",
    "BKK_056226": "Fővám tér",
    "BKK_056228": "Kálvin tér",
    "BKK_056230": "Rákóczi tér",
    "BKK_056232": "II. János Pál pápa tér",
    "BKK_056234": "Keleti pályaudvar",
    "BKK_09001188": "Batthyány tér",
    "BKK_09019191": "Margit híd",
    "BKK_09043193": "Szépvölgyi út",
    "BKK_09050195": "Tímár utca",
    "BKK_09068197": "Szentlélek tér",
    "BKK_09118199": "Filatorigát",
    "BKK_09084201": "Kaszásdűlű",
    "BKK_09100203": "Aquincum",
    "BKK_09159205": "Rómaifürdő",
    "BKK_19720237": "Közvágóhíd",
    "BKK_19726239": "Kén utca",
    "BKK_19729240": "Pesterzsébet felső",
    "BKK_09220225": "Boráros tér",
    "BKK_09221227": "Müpa",
    "BKK_09223228": "Szabadkikötő",
    "BKK_19795279": "Örs vezér tere",
    "BKK_19798282": "Rákosfalva",
    "BKK_09001189": "Batthyány tér",
    "BKK_19795280": "Örs vezér tere",
}


@app.route("/schedules", defaults={"route_id": None}, methods=["GET"])
@app.route("/schedules/<route_id>", methods=["GET"])
def get_schedules(route_id: str | None) -> str:
    """Return HTML page with the filtered schedules.

    :param route_id: Specify a route_id to filter the jobs. Use None the return
        all jobs.
    """
    from BudapestMetroDisplay.bkk_opendata import departure_scheduler
    from BudapestMetroDisplay.led_control import ROUTE_COLORS

    jobs = departure_scheduler.get_jobs()
    job_list = []
    for job in jobs:
        if route_id is None or job.args[1] == route_id:
            # Add the job to the list if route_id wasn't specified,
            # or if it matches the route_id of the job
            job_info = {
                "id": job.id,
                "stop_name": stop_names.get(job.args[0], "Unknown"),
                "arg1": job.args[1],
                "arg2": job.args[2],
                "arg3": job.args[3],
                "arg4": job.args[4],
            }
            job_list.append(job_info)
    return render_template("schedules.html", jobs=job_list, route_colors=ROUTE_COLORS)


def start_webserver() -> None:
    """Start the webserver in a separate thread."""
    thread = threading.Thread(
        target=lambda: app.run(debug=False, use_reloader=False),
        daemon=True,
        name="Webserver thread",
    )
    thread.start()



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

from typing import Any

import numpy as np
from apscheduler.job import Job
from apscheduler.schedulers.base import BaseScheduler


def count_jobs_by_argument(
    scheduler: BaseScheduler,
    search_value: Any,
    arg_position: int,
) -> int:
    """Count the number of jobs in a BaseScheduler which matches the search value.

    :param scheduler: BackgroundScheduler instance
    :param search_value: The value to match the argument with
    :param arg_position: The index of the argument we want to compare
    :return: The number of jobs in which the argument matches the search value
    """
    i = 0
    for job in scheduler.get_jobs():
        job_args = job.args
        if len(job_args) > arg_position and job_args[arg_position] == search_value:
            i += 1
    return i


def get_jobs_by_argument(
    scheduler: BaseScheduler,
    search_value: Any,
    arg_position: int,
) -> list[Job]:
    """Return the jobs from a BaseScheduler which arguments matches the search value.

    :param scheduler: BackgroundScheduler instance
    :param search_value: The value to match the argument with
    :param arg_position: The index of the argument we want to compare
    :return: The jobs that are matches the search value
    """
    jobs = scheduler.get_jobs()
    filtered_jobs = []

    for job in jobs:
        job_args = job.args
        # Check if the argument matches
        if len(job_args) > arg_position and job.args[arg_position] == search_value:
            filtered_jobs.append(job)

    return filtered_jobs


def find_soonest_job_by_argument(
    scheduler: BaseScheduler,
    search_value: Any,
    arg_position: int,
) -> Job | None:
    """Find the soonest job based on the next run time, filtered by a specific argument.

    :param scheduler: The APScheduler instance.
    :param search_value: The value to match the argument with
    :param arg_position: The index of the argument we want to compare
    :return: The job with the soonest schedule time
    """
    soonest_job = None
    for job in scheduler.get_jobs():
        job_args = job.args
        # Check if the job has the specified argument and value
        if (
            len(job_args) > arg_position
            and job_args[arg_position] == search_value
            and (
                soonest_job is None
                or (job.next_run_time and job.next_run_time < soonest_job.next_run_time)
            )
        ):
            soonest_job = job

    return soonest_job


def calculate_average_time_between_jobs(filtered_jobs: list[Job]) -> float | None:
    """Calculate the average time between the jobs supplied in a list.

    :param filtered_jobs: A list of Jobs to check
    :return: Return the average time between jobs in seconds,
        or None, when the time cannot be determined
    """
    # Extract the next run times and sort them
    run_times = [job.next_run_time for job in filtered_jobs if job.next_run_time]
    run_times = sorted(run_times)

    if len(run_times) < 2:
        return None  # Not enough jobs to calculate time difference

    # Calculate differences in seconds
    time_differences = [
        (run_times[i] - run_times[i - 1]).total_seconds()
        for i in range(1, len(run_times))
    ]

    # Calculate average difference
    return np.mean(time_differences) if time_differences else None



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import argparse
import asyncio
import logging
import signal
import sys
import time
from asyncio import AbstractEventLoop

from BudapestMetroDisplay import bkk_opendata, led_control, log, webserver
from BudapestMetroDisplay._version import __version__
from BudapestMetroDisplay.config import settings
from BudapestMetroDisplay.stops import alert_routes, stops_metro, stops_railway

if settings.esphome.used:
    from BudapestMetroDisplay.esphome import (
        connect_and_subscribe,
        start_background_loop,
    )

logger = logging.getLogger(__name__)

parser = None
loop: AbstractEventLoop


def handle_exit_signal(_signum, _frame) -> None:  # noqa: ANN001
    """Handle signals for a clean exit."""
    logger.info("Signal received, stopping threads...")
    bkk_opendata.departure_scheduler.shutdown()
    logger.debug("Departure scheduler shut down")
    bkk_opendata.led_scheduler.shutdown(wait=False)
    logger.debug("LED scheduler shut down")
    bkk_opendata.api_update_scheduler.shutdown(wait=False)
    logger.debug("API Update scheduler shut down")
    led_control.deactivate_sacn()
    logger.debug("sACN update thread shut down")
    if settings.esphome.used:
        loop.stop()
        logger.debug("ESPHome update thread shut down")
    logger.info("Cleanup complete. Exiting...")
    sys.exit(0)


# Set the global exception hook
sys.excepthook = log.log_exception


def main() -> None:  # noqa: D103
    global parser, loop

    if settings.esphome.used:
        loop = start_background_loop()
        loop.call_soon_threadsafe(asyncio.create_task, connect_and_subscribe())

    # Set up argument parser
    parser = argparse.ArgumentParser(
        description="Run the BKK Opendata and LED Control Program.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug mode for verbose output.",
    )
    parser.add_argument(
        "--trace",
        action="store_true",
        help="Enable trace mode for verbose output.",
    )

    # Set up logging with or without debug mode
    log.setup_logging(parser)
    logger.info(f"Program started, version {__version__}")

    # Register the exit handler for signals (e.g., Ctrl+C)
    signal.signal(signal.SIGINT, handle_exit_signal)
    signal.signal(signal.SIGTERM, handle_exit_signal)

    # Create schedules for updating the departures data
    bkk_opendata.create_schedule_updates(stops_metro, "REGULAR")
    bkk_opendata.create_schedule_updates(stops_railway, "REGULAR")
    # Create schedules for updating the realtime data
    bkk_opendata.create_schedule_updates(stops_railway, "REALTIME")
    # Create schedules for updating the alarm data for non-realtime stops
    bkk_opendata.create_alert_updates(alert_routes)

    # Start the sACN sending routine with continuous updates
    led_control.reset_leds_to_default()
    # Start sending LED data via sACN
    led_control.activate_sacn()

    webserver.start_webserver()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#  MIT License
#
#  Copyright (c) 2024 denes44
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import logging
import math
import tkinter as tk
from pathlib import Path
from tkinter import ttk

from apscheduler.schedulers.background import BackgroundScheduler

from BudapestMetroDisplay import bkk_opendata

logger = logging.getLogger(__name__)


def on_closing() -> None:
    """Handle the GUI window close event."""
    logger.info("GUI window closed, stopping application...")
    root.destroy()


# Create the main window
root = tk.Tk()
root.title("BudapestMetroDisplay LED Test GUI")
root.protocol("WM_DELETE_WINDOW", on_closing)

# Create a Canvas to draw on
canvas = tk.Canvas(root, width=700, height=1000)
canvas.pack()

# Background image
background_image_path = "../../resources/gui_bg.png"
if Path(background_image_path).exists():
    background_image = tk.PhotoImage(file=background_image_path)
    # Set the image as the background on the canvas
    canvas.create_image(0, 0, anchor=tk.NW, image=background_image)

# Example positions for LEDs (list of tuples: (x, y))
led_positions = [
    (418.06, 960.43, -3),  # (376.26 * 1.1111, 864.39 * 1.1111, -3)
    (380.00, 848.43, -25),  # (342.00 * 1.1111, 763.59 * 1.1111, -25)
    (345.91, 780.33, -28),  # (310.32 * 1.1111, 702.30 * 1.1111, -28)
    (308.60, 894.78, -2),  # (277.74 * 1.1111, 805.50 * 1.1111, -2)
    (305.23, 752.33, -7),  # (274.71 * 1.1111, 677.70 * 1.1111, -7)
    (290.93, 698.94, -25),  # (261.24 * 1.1111, 628.05 * 1.1111, -25)
    (46.86, 810.36, -346),  # (42.18 * 1.1111, 729.33 * 1.1111, -346)
    (101.56, 802.86, -22),  # (91.41 * 1.1111, 722.58 * 1.1111, -22)
    (164.00, 737.57, -79),  # (147.60 * 1.1111, 663.81 * 1.1111, -79)
    (170.43, 713.11, -73),  # (153.39 * 1.1111, 641.70 * 1.1111, -73)
    (204.13, 675.03, -58),  # (183.72 * 1.1111, 607.53 * 1.1111, -58)
    (219.99, 651.00, -53),  # (197.49 * 1.1111, 585.90 * 1.1111, -53)
    (243.00, 628.83, -55),  # (218.70 * 1.1111, 565.95 * 1.1111, -55)
    (289.43, 603.56, -14),  # (260.49 * 1.1111, 543.21 * 1.1111, -14)
    (317.43, 582.60, -340),  # (285.69 * 1.1111, 524.70 * 1.1111, -340)
    (60.10, 543.24, -90),  # (54.09 * 1.1111, 489.81 * 1.1111, -90)
    (60.07, 502.76, -58),  # (54.06 * 1.1111, 452.49 * 1.1111, -58)
    (126.21, 504.40, -340),  # (114.09 * 1.1111, 453.96 * 1.1111, -340)
    (164.43, 511.73, -354),  # (147.99 * 1.1111, 460.56 * 1.1111, -354)
    (204.35, 568.00, -270),  # (183.45 * 1.1111, 511.20 * 1.1111, -270)
    (231.72, 592.00, -5),  # (208.53 * 1.1111, 532.80 * 1.1111, -5)
    (284.00, 571.00, -23),  # (255.60 * 1.1111, 514.80 * 1.1111, -23)
    (326.40, 554.12, -15),  # (293.76 * 1.1111, 499.74 * 1.1111, -15)
    (457.96, 550.43, -9),  # (411.15 * 1.1111, 495.39 * 1.1111, -9)
    (524.00, 543.00, -8),  # (471.60 * 1.1111, 488.70 * 1.1111, -8)
    (600.00, 530.76, -9),  # (540.00 * 1.1111, 477.57 * 1.1111, -9)
    (668.34, 518.94, -10),  # (602.10 * 1.1111, 467.04 * 1.1111, -10)
    (127.66, 454.77, -2),  # (114.90 * 1.1111, 409.29 * 1.1111, -2)
    (136.32, 358.90, -342),  # (122.79 * 1.1111, 323.01 * 1.1111, -342)
    (152.50, 325.54, -323),  # (137.25 * 1.1111, 293.19 * 1.1111, -323)
    (162.33, 285.92, -2),  # (146.10 * 1.1111, 257.31 * 1.1111, -2)
    (161.91, 208.88, -30),  # (145.71 * 1.1111, 186.99 * 1.1111, -30)
    (155.51, 169.87, -340),  # (139.65 * 1.1111, 151.89 * 1.1111, -340)
    (168.77, 99.99, -25),  # (151.59 * 1.1111, 89.19 * 1.1111, -25)
    (160.45, 48.83, -359),  # (144.15 * 1.1111, 43.50 * 1.1111, -359)
    (372.00, 117.71, -176),  # (336.00 * 1.1111, 105.54 * 1.1111, -176)
    (326.43, 126.00, -210),  # (293.79 * 1.1111, 113.40 * 1.1111, -210)
    (296.60, 198.77, -255),  # (266.91 * 1.1111, 178.59 * 1.1111, -255)
    (276.00, 273.98, -254),  # (248.40 * 1.1111, 246.60 * 1.1111, -254)
    (264.16, 316.54, -254),  # (237.84 * 1.1111, 284.79 * 1.1111, -254)
    (247.17, 374.21, -253),  # (222.54 * 1.1111, 336.75 * 1.1111, -253)
    (231.10, 426.43, -255),  # (208.29 * 1.1111, 383.79 * 1.1111, -255)
    (209.17, 476.00, -253),  # (188.25 * 1.1111, 428.40 * 1.1111, -253)
    (205.74, 532.06, -268),  # (185.16 * 1.1111, 478.65 * 1.1111, -268)
    (212.13, 600.93, -289),  # (190.92 * 1.1111, 540.84 * 1.1111, -289)
    (279.79, 652.03, -330),  # (251.82 * 1.1111, 587.73 * 1.1111, -330)
    (324.00, 677.56, -333),  # (291.60 * 1.1111, 609.15 * 1.1111, -333)
    (372.36, 702.26, -332),  # (335.70 * 1.1111, 632.04 * 1.1111, -332)
    (419.33, 727.78, -332),  # (377.40 * 1.1111, 654.00 * 1.1111, -332)
    (480.43, 760.00, -331),  # (432.39 * 1.1111, 684.00 * 1.1111, -331)
    (510.00, 775.31, -331),  # (459.00 * 1.1111, 698.79 * 1.1111, -331)
    (552.16, 798.82, -338),  # (496.95 * 1.1111, 718.95 * 1.1111, -338)
    (663.08, 811.83, -335),  # (597.03 * 1.1111, 730.65 * 1.1111, -335)
    (186.00, 571.52, -30),  # (168.00 * 1.1111, 515.37 * 1.1111, -30)
    (210.21, 550.66, -44),  # (189.81 * 1.1111, 495.60 * 1.1111, -44)
    (226.56, 535.76, -45),  # (203.82 * 1.1111, 482.13 * 1.1111, -45)
    (247.00, 515.00, -45),  # (222.30 * 1.1111, 464.40 * 1.1111, -45)
    (262.42, 500.56, -45),  # (236.19 * 1.1111, 450.51 * 1.1111, -45)
    (275.38, 486.56, -45),  # (248.46 * 1.1111, 437.91 * 1.1111, -45)
    (292.43, 470.33, -45),  # (263.19 * 1.1111, 423.00 * 1.1111, -45)
    (313.50, 449.33, -45),  # (282.15 * 1.1111, 404.40 * 1.1111, -45)
    (335.81, 425.88, -358),  # (301.23 * 1.1111, 383.49 * 1.1111, -358)
    (388.16, 414.16, -347),  # (349.35 * 1.1111, 371.73 * 1.1111, -347)
]
stop_names = {
    "BKK_F00965": "Vörösmarty tér",
    "BKK_F00963": "Deák Ferenc tér",
    "BKK_F00997": "Bajcsy-Zsilinszky út",
    "BKK_F01080": "Opera",
    "BKK_F01086": "Oktogon",
    "BKK_F01093": "Vörösmarty utca",
    "BKK_F01096": "Kodály körönd",
    "BKK_F01101": "Bajza utca",
    "BKK_F01103": "Hősök tere",
    "BKK_F02697": "Széchenyi fürdő",
    "BKK_F02888": "Mexikói út",
    "BKK_F00094": "Déli pályaudvar",
    "BKK_F02481": "Szél Kálmán tér",
    "BKK_F00063": "Batthyány tér",
    "BKK_F00959": "Kossuth Lajos tér",
    "BKK_F00961": "Deák Ferenc tér",
    "BKK_F01019": "Astoria",
    "BKK_F01292": "Blaha Lujza tér",
    "BKK_F01336": "Keleti pályaudvar",
    "BKK_F01325": "Puskás Ferenc Stadion",
    "BKK_F01743": "Pillangó utca",
    "BKK_F01749": "Örs vezér tere",
    "BKK_F00913": "Újpest-központ",
    "BKK_F00898": "Újpest-városkapu",
    "BKK_F02685": "Gyöngyösi utca",
    "BKK_F02683": "Forgách utca",
    "BKK_F02638": "Göncz Árpád városközpont",
    "BKK_F02681": "Dózsa György út",
    "BKK_F02614": "Lehel tér",
    "BKK_F00937": "Nyugati pályaudvar",
    "BKK_F00957": "Arany János utca",
    "BKK_F00955": "Deák Ferenc tér",
    "BKK_F00953": "Ferenciek tere",
    "BKK_F01290": "Kálvin tér",
    "BKK_F01189": "Corvin-negyed",
    "BKK_F01233": "Klinikák",
    "BKK_F01253": "Nagyvárad tér",
    "BKK_F01283": "Népliget",
    "BKK_F01494": "Ecseri út",
    "BKK_F01879": "Pöttyös utca",
    "BKK_F01542": "Határ út",
    "BKK_F01544": "Kőbánya-Kispest",
    "BKK_056215": "Kelenföldi pályaudvar",
    "BKK_056217": "Bikás park",
    "BKK_056219": "Újbuda-Központ",
    "BKK_056221": "Móricz Zsigmond körtér",
    "BKK_056223": "Szent Gellért tér",
    "BKK_056225": "Fővám tér",
    "BKK_056227": "Kálvin tér",
    "BKK_056229": "Rákóczi tér",
    "BKK_056231": "II. János Pál pápa tér",
    "BKK_056233": "Keleti pályaudvar",
    "BKK_09001187": "Batthyány tér",
    "BKK_09019190": "Margit híd",
    "BKK_09043192": "Szépvölgyi út",
    "BKK_09050194": "Tímár utca",
    "BKK_09068196": "Szentlélek tér",
    "BKK_09118198": "Filatorigát",
    "BKK_09084200": "Kaszásdűlű",
    "BKK_09100202": "Aquincum",
    "BKK_09159204": "Rómaifürdő",
    "BKK_19720236": "Közvágóhíd",
    "BKK_19726238": "Kén utca",
    "BKK_19729241": "Pesterzsébet felső",
    "BKK_09220224": "Boráros tér",
    "BKK_09221226": "Müpa",
    "BKK_09223229": "Szabadkikötő",
    "BKK_19795278": "Örs vezér tere",
    "BKK_19798281": "Rákosfalva",
    "BKK_F00964": "Vörösmarty tér",
    "BKK_F00962": "Deák Ferenc tér",
    "BKK_F00996": "Bajcsy-Zsilinszky út",
    "BKK_F01079": "Opera",
    "BKK_F01085": "Oktogon",
    "BKK_F01092": "Vörösmarty utca",
    "BKK_F01095": "Kodály körönd",
    "BKK_F01100": "Bajza utca",
    "BKK_F01102": "Hősök tere",
    "BKK_F02696": "Széchenyi fürdő",
    "BKK_F02887": "Mexikói út",
    "BKK_F00093": "Déli pályaudvar",
    "BKK_F02480": "Szél Kálmán tér",
    "BKK_F00062": "Batthyány tér",
    "BKK_F00958": "Kossuth Lajos tér",
    "BKK_F00960": "Deák Ferenc tér",
    "BKK_F01018": "Astoria",
    "BKK_F01291": "Blaha Lujza tér",
    "BKK_F01335": "Keleti pályaudvar",
    "BKK_F01324": "Puskás Ferenc Stadion",
    "BKK_F01742": "Pillangó utca",
    "BKK_F00912": "Újpest-központ",
    "BKK_F00897": "Újpest-városkapu",
    "BKK_F02684": "Gyöngyösi utca",
    "BKK_F02682": "Forgách utca",
    "BKK_F02637": "Göncz Árpád városközpont",
    "BKK_F02680": "Dózsa György út",
    "BKK_F02613": "Lehel tér",
    "BKK_F00936": "Nyugati pályaudvar",
    "BKK_F00956": "Arany János utca",
    "BKK_F00954": "Deák Ferenc tér",
    "BKK_F00952": "Ferenciek tere",
    "BKK_F01289": "Kálvin tér",
    "BKK_F01188": "Corvin-negyed",
    "BKK_F01232": "Klinikák",
    "BKK_F01252": "Nagyvárad tér",
    "BKK_F01282": "Népliget",
    "BKK_F01493": "Ecseri út",
    "BKK_F01878": "Pöttyös utca",
    "BKK_F01541": "Határ út",
    "BKK_F01543": "Kőbánya-Kispest",
    "BKK_056216": "Kelenföldi pályaudvar",
    "BKK_056218": "Bikás park",
    "BKK_056220": "Újbuda-Központ",
    "BKK_056222": "Móricz Zsigmond körtér",
    "BKK_056224": "Szent Gellért tér",
    "BKK_056226": "Fővám tér",
    "BKK_056228": "Kálvin tér",
    "BKK_056230": "Rákóczi tér",
    "BKK_056232": "II. János Pál pápa tér",
    "BKK_056234": "Keleti pályaudvar",
    "BKK_09001188": "Batthyány tér",
    "BKK_09019191": "Margit híd",
    "BKK_09043193": "Szépvölgyi út",
    "BKK_09050195": "Tímár utca",
    "BKK_09068197": "Szentlélek tér",
    "BKK_09118199": "Filatorigát",
    "BKK_09084201": "Kaszásdűlű",
    "BKK_09100203": "Aquincum",
    "BKK_09159205": "Rómaifürdő",
    "BKK_19720237": "Közvágóhíd",
    "BKK_19726239": "Kén utca",
    "BKK_19729240": "Pesterzsébet felső",
    "BKK_09220225": "Boráros tér",
    "BKK_09221227": "Müpa",
    "BKK_09223228": "Szabadkikötő",
    "BKK_19795279": "Örs vezér tere",
    "BKK_19798282": "Rákosfalva",
    "BKK_09001189": "Batthyány tér",
    "BKK_19795280": "Örs vezér tere",
}

led_rectangles: list[int | None] = [None] * len(led_positions)

# Width and height of the rectangles (LEDs)
RECT_WIDTH = 10
RECT_HEIGHT = 10


# Function to convert the LED's RGB state to a color string
def rgb_to_color_string(color: tuple[int, int, int]) -> str:
    """Convert RGB tuple values to hex format.

    :param color: An RGB color value as a tuple
    :return: Color value converted to hex string
    """
    return f"#{color[0]:02x}{color[1]:02x}{color[2]:02x}"


def get_rotated_rectangle(
    x: float,
    y: float,
    width: float,
    height: float,
    rotation: float,
) -> list[tuple[float, float]]:
    """Get the rotated points of a rectangle."""
    # Calculate half-width and half-height
    hw, hh = width / 2, height / 2

    # Define the four corners of the rectangle (relative to center)
    corners = [(-hw, -hh), (hw, -hh), (hw, hh), (-hw, hh)]

    # Convert rotation from degrees to radians
    angle_rad = math.radians(rotation)

    # Apply rotation to each corner using a rotation matrix
    return [
        (
            x + (corner[0] * math.cos(angle_rad) - corner[1] * math.sin(angle_rad)),
            y + (corner[0] * math.sin(angle_rad) + corner[1] * math.cos(angle_rad)),
        )
        for corner in corners
    ]


# Function to draw the LEDs as circles
def draw_leds() -> None:
    """Draw the LEDs on the canvas."""
    from BudapestMetroDisplay.led_control import get_led_color

    for i, (x, y, rotation) in enumerate(led_positions):
        # Get the color from led_states (convert from flat list to RGB)
        led_color = get_led_color(i)
        color = rgb_to_color_string(
            led_color if not None else (0, 0, 0),  # type: ignore[arg-type]
        )

        # Get the rotated rectangle coordinates
        points = get_rotated_rectangle(x, y, RECT_WIDTH, RECT_HEIGHT, rotation)

        # Draw the rectangle with rotation
        led_rectangles[i] = canvas.create_polygon(points, fill=color, outline="white")
    logger.debug("Drawing of the LED on the GUI is finished")


def change_gui_led_color(led_index: int, color: tuple[int, int, int]) -> None:
    """Change the color of a LED on the canvas."""
    led_rectangle = led_rectangles[led_index]
    if led_rectangle is not None:
        canvas.itemconfig(led_rectangle, fill=rgb_to_color_string(color))


def filter_jobs_by_route(
    job_table: ttk.Treeview,
    scheduler: BackgroundScheduler,
    route_id: str,
) -> None:
    """Filter and update job table with jobs for the given route_id."""
    for row in job_table.get_children():
        job_table.delete(row)

    for job in scheduler.get_jobs():
        args = job.args or (None, None, None, None, None)
        if args[1] == route_id:  # Check if the route_id matches
            job_table.insert(
                "",
                "end",
                values=(
                    job.id,
                    job.name,
                    job.next_run_time,
                    job.trigger,
                    stop_names[str(args[0])],
                    args[1],
                    args[2],
                    args[3],
                    args[4],
                ),
            )


def create_filtered_job_table(scheduler: BackgroundScheduler, route_id: str) -> None:
    """Create a job schedule window filtered by route_id."""
    job_window = tk.Toplevel(root)
    job_window.title(f"Job Schedule for {route_id}")

    columns = (
        "ID",
        "Name",
        "Next Run Time",
        "Trigger",
        "stop_id",
        "route_id",
        "trip_id",
        "job_time",
        "delay",
    )
    job_table = ttk.Treeview(job_window, columns=columns, show="headings")
    job_table.pack(fill=tk.BOTH, expand=True)

    for col in columns:
        job_table.heading(col, text=col)
        job_table.column(col, width=100, anchor=tk.CENTER)

    filter_jobs_by_route(job_table, scheduler, route_id)


def create_route_buttons(my_canvas: tk.Canvas) -> None:
    """Create buttons for each route_id in ROUTE_COLORS."""
    from BudapestMetroDisplay.led_control import ROUTE_COLORS

    row, col = 0, 0
    x_offset, y_offset = (
        10,
        10,
    )  # Offset to position buttons with some space from the top-left corner
    for route_id, color in ROUTE_COLORS.items():
        color_hex = rgb_to_color_string(color)
        btn = tk.Button(
            my_canvas,
            text=route_id,
            bg=color_hex,
            command=lambda rid=route_id:  # type: ignore[misc]
            create_filtered_job_table(bkk_opendata.departure_scheduler, rid),
        )
        my_canvas.create_window(
            x_offset + col * 110,
            y_offset + row * 40,
            window=btn,
            anchor="nw",
        )  # Positioning buttons
        col += 1
        if col == 1:  # Adjust for buttons per row
            col = 0
            row += 1


def start_gui() -> None:
    """Open the GUI window and draw the elements."""
    draw_leds()
    create_route_buttons(canvas)
    root.mainloop()
    logger.debug("GUI started")



#  MIT License
#
#  Copyright (c) 2024 denes44
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

# ruff: noqa: D103, S101

import os

import pytest
from pydantic import ValidationError

# Mock environment variables before importing the module
# LED Configuration
os.environ["LED_DIM_RATIO"] = "0.3"
os.environ["LED_FADE_TIME"] = "5.3"

# sACN Configuration
os.environ["SACN_MULTICAST"] = "False"
os.environ["SACN_UNICAST_IP"] = "192.168.1.1"
os.environ["SACN_UNIVERSE"] = "2"
os.environ["SACN_FPS"] = "15"

# BKK Configuration
os.environ["BKK_API_KEY"] = "123e4567-e89b-12d3-a456-426614174000"
os.environ["BKK_API_UPDATE_INTERVAL"] = "6"
os.environ["BKK_API_UPDATE_REALTIME"] = "82"
os.environ["BKK_API_UPDATE_REGULAR"] = "3456"
os.environ["BKK_API_UPDATE_ALERTS"] = "123"

# ESPHome Configuration
os.environ["ESPHOME_USED"] = "True"
os.environ["ESPHOME_DEVICE_IP"] = "192.168.1.2"
os.environ["ESPHOME_API_KEY"] = "0LTLKmoTVR0BO3xppXQkIBVb0VzDLZFqAplYnADTbOY="

from BudapestMetroDisplay.config import (
    AppConfig,
    BKKConfig,
    ESPHomeConfig,
    LEDConfig,
    SACNConfig,
)


def test_app_config_initializes_correctly() -> None:
    config = AppConfig()
    # LED Configuration
    assert config.led.dim_ratio == 0.3
    assert config.led.fade_time == 5.3

    # sACN Configuration
    assert not config.sacn.multicast
    assert str(config.sacn.unicast_ip) == "192.168.1.1"
    assert config.sacn.universe == 2
    assert config.sacn.fps == 15

    # BKK Configuration
    assert config.bkk.api_key == "123e4567-e89b-12d3-a456-426614174000"
    assert config.bkk.api_update_interval == 6
    assert config.bkk.api_update_realtime == 82
    assert config.bkk.api_update_regular == 3456
    assert config.bkk.api_update_alerts == 123

    # ESPHome Configuration
    assert config.esphome.used
    assert str(config.esphome.device_ip) == "192.168.1.2"
    assert config.esphome.api_key == "0LTLKmoTVR0BO3xppXQkIBVb0VzDLZFqAplYnADTbOY="


def test_delete_env_vars() -> None:
    # List of environment variable keys to delete
    env_vars = [
        "LED_DIM_RATIO",
        "LED_FADE_TIME",
        "SACN_MULTICAST",
        "SACN_UNICAST_IP",
        "SACN_UNIVERSE",
        "SACN_FPS",
        "BKK_API_KEY",
        "BKK_API_UPDATE_INTERVAL",
        "BKK_API_UPDATE_REALTIME",
        "BKK_API_UPDATE_REGULAR",
        "BKK_API_UPDATE_ALERTS",
        "ESPHOME_USED",
        "ESPHOME_DEVICE_IP",
        "ESPHOME_API_KEY",
    ]
    # Delete each environment variable
    for var in env_vars:
        if var in os.environ:
            del os.environ[var]


def test_led_config_dim_ratio_within_bounds() -> None:
    config = LEDConfig(dim_ratio=0.5)
    assert config.dim_ratio == 0.5


def test_led_config_dim_ratio_out_of_bounds_positive() -> None:
    with pytest.raises(ValidationError):
        LEDConfig(dim_ratio=1.5)


def test_led_config_dim_ratio_out_of_bounds_negative() -> None:
    with pytest.raises(ValidationError):
        LEDConfig(dim_ratio=-2.5)


def test_led_config_fade_time_positive() -> None:
    config = LEDConfig(fade_time=2.0)
    assert config.fade_time == 2.0


def test_led_config_fade_time_zero() -> None:
    config = LEDConfig(fade_time=0.0)
    assert config.fade_time == 0.0


def test_led_config_fade_time_non_positive() -> None:
    with pytest.raises(ValidationError):
        LEDConfig(fade_time=-1.0)


def test_sacn_config_multicast_default() -> None:
    config = SACNConfig()
    assert config.multicast is True


def test_sacn_config_unicast_ip_required() -> None:
    with pytest.raises(ValidationError):
        SACNConfig(multicast=False, unicast_ip=None)


def test_sacn_config_unicast_ipv4() -> None:
    config = SACNConfig(multicast=False, unicast_ip="192.168.1.1")
    assert str(config.unicast_ip) == "192.168.1.1"


def test_sacn_config_unicast_ipv6() -> None:
    config = SACNConfig(
        multicast=False,
        unicast_ip="2001:0000:130F:0000:0000:09C0:876A:130B",
    )
    assert str(config.unicast_ip) == "2001:0:130f::9c0:876a:130b"


def test_bkk_config_api_key_required_none() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_key=None)


def test_bkk_config_api_key_required_empty() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_key="")


def test_bkk_config_api_key_required_invalid() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_key="test-api-key")


def test_bkk_config_api_key_required() -> None:
    config = BKKConfig(api_key="123e4567-e89b-12d3-a456-426614174000")
    assert config.api_key == "123e4567-e89b-12d3-a456-426614174000"


def test_bkk_config_api_update_interval_positive() -> None:
    config = BKKConfig(
        api_key="123e4567-e89b-12d3-a456-426614174000",
        api_update_interval=5,
    )
    assert config.api_update_interval == 5


def test_bkk_config_api_update_interval_out_of_bounds() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_update_interval=-5)


def test_bkk_config_api_update_realtime_positive() -> None:
    config = BKKConfig(
        api_key="123e4567-e89b-12d3-a456-426614174000",
        api_update_realtime=5,
    )
    assert config.api_update_realtime == 5


def test_bkk_config_api_update_realtime_out_of_bounds() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_update_realtime=-5)


def test_bkk_config_api_update_regular_positive() -> None:
    config = BKKConfig(
        api_key="123e4567-e89b-12d3-a456-426614174000",
        api_update_regular=5,
    )
    assert config.api_update_regular == 5


def test_bkk_config_api_update_regular_out_of_bounds() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_update_regular=-5)


def test_bkk_config_api_update_alerts_positive() -> None:
    config = BKKConfig(
        api_key="123e4567-e89b-12d3-a456-426614174000",
        api_update_alerts=5,
    )
    assert config.api_update_alerts == 5


def test_bkk_config_api_update_alerts_out_of_bounds() -> None:
    with pytest.raises(ValidationError):
        BKKConfig(api_update_alerts=-5)


def test_esphome_config_used_requires_ip() -> None:
    with pytest.raises(ValidationError):
        ESPHomeConfig(used=True, device_ip=None, api_key=None)


def test_esphome_config_used_requires_ip_and_key() -> None:
    with pytest.raises(ValidationError):
        ESPHomeConfig(used=True, device_ip=None, api_key=None)


def test_esphome_config_used_empty_api_key() -> None:
    with pytest.raises(ValidationError):
        ESPHomeConfig(used=True, device_ip="192.168.1.1", api_key="")


def test_esphome_config_used_invalid_api_key() -> None:
    with pytest.raises(ValidationError):
        ESPHomeConfig(used=True, device_ip="192.168.1.1", api_key="test_api_key")


def test_esphome_config_used_ipv6() -> None:
    config = ESPHomeConfig(
        used=True,
        device_ip="2001:0000:130F:0000:0000:09C0:876A:130B",
        api_key="0LTLKmoTVR0BO3xppXQkIBVb0VzDLZFqAplYnADTbOY=",
    )
    assert config.used is True
    assert str(config.device_ip) == "2001:0:130f::9c0:876a:130b"
    assert config.api_key == "0LTLKmoTVR0BO3xppXQkIBVb0VzDLZFqAplYnADTbOY="


def test_esphome_config_not_used_allows_missing_ip_and_key() -> None:
    config = ESPHomeConfig(used=False)
    assert config.used is False
    assert config.device_ip is None
    assert config.api_key is None



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.
import argparse
import logging
from datetime import datetime
from logging.handlers import RotatingFileHandler
from pathlib import Path

from BudapestMetroDisplay.config import settings

logger = logging.getLogger(__name__)

# Define the custom log level
TRACE_LEVEL = 5


def add_logging_level(
    level_name: str,
    level_num: int,
    method_name: str | None = None,
) -> None:
    """Add a new logging level to the `logging` module.

    `level_name` becomes an attribute of the `logging` module with the value
    `level_num`. `method_name` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `method_name` is not specified, `level_name.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example:
    -------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        msg = f"{level_name} already defined in logging module"
        raise AttributeError(msg)
    if hasattr(logging, method_name):
        msg = f"{method_name} already defined in logging module"
        raise AttributeError(msg)
    if hasattr(logging.getLoggerClass(), method_name):
        msg = f"{method_name} already defined in logger class"
        raise AttributeError(msg)

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(
        self,  # noqa: ANN001
        message,  # noqa: ANN001
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> None:
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(
        message,  # noqa: ANN001
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> None:
        logging.log(level_num, message, *args, **kwargs)  # noqa: LOG015,RUF100

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def setup_logging(parser: argparse.ArgumentParser) -> None:
    """Init the logging interfaces for the application (file, STD)."""
    global logger

    # Define custom log level
    add_logging_level("TRACE", logging.DEBUG - 5)

    # Create logger instance
    logger = logging.getLogger()

    # Parse command-line arguments
    args = parser.parse_args()
    if args.trace:
        logger.setLevel(logging.TRACE)  # type: ignore[attr-defined]
    elif args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # Create a rotating file handler
    file_handler = RotatingFileHandler(
        Path(settings.log.path) / "application.log",
        maxBytes=5 * 1024 * 1024,  # 5 MB file size limit
        backupCount=10,  # Keep up to 10 backup log files
    )

    # Create a rotating file handler
    file_handler_error = RotatingFileHandler(
        Path(settings.log.path) / "application_error.log",
        maxBytes=5 * 1024 * 1024,  # 5 MB file size limit
        backupCount=5,  # Keep up to 5 backup log files
    )

    # Create a temporary logger for adding the start entry
    start_logger = logging.getLogger("start_logger")
    start_logger.setLevel(logging.INFO)  # Temporarily allow INFO for this logger
    # Add handler to the logger
    start_logger.addHandler(file_handler)
    start_logger.addHandler(file_handler_error)
    # Add a custom log entry at program start to separate runs
    separator = "=" * 50
    start_logger.info(separator)
    start_logger.info(f"Program started at {datetime.now()}")
    start_logger.info(separator)
    # Remove the handler after logging the start messages
    start_logger.removeHandler(file_handler)
    start_logger.removeHandler(file_handler_error)

    # Capture only WARNING and higher
    file_handler_error.setLevel(logging.WARNING)

    # Create a console handler (prints to stdout)
    console_handler = logging.StreamHandler()

    # Create a logging format
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(funcName)s: %(message)s",
    )
    file_handler.setFormatter(formatter)
    file_handler_error.setFormatter(formatter)
    console_handler.setFormatter(formatter)

    # Add handlers to the logger (without using basicConfig())
    if not logger.hasHandlers():  # Prevent adding handlers multiple times
        logger.addHandler(file_handler)
        logger.addHandler(file_handler_error)
        logger.addHandler(console_handler)

    if logger.getEffectiveLevel() == logging.TRACE:  # type: ignore[attr-defined]
        logger.trace("Trace level logging enabled")  # type: ignore[attr-defined]
    elif logger.getEffectiveLevel() == logging.DEBUG:
        logger.debug("Debug level logging enabled")


def log_exception(exc_type, exc_value, exc_traceback) -> None:  # noqa: ANN001
    """Log an exception."""
    if issubclass(exc_type, KeyboardInterrupt):
        # Don't log KeyboardInterrupt exceptions
        logger.info("Program interrupted by user, exiting...")
        return
    # Log the exception details
    logger.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))



#  MIT License
#
#  Copyright (c) 2024 denes44
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import asyncio
import logging
import threading
from asyncio import AbstractEventLoop

from aioesphomeapi import (
    APIClient,
    APIConnectionError,
    EntityState,
    LightState,
    ReconnectLogic,
)

from BudapestMetroDisplay._version import __version__
from BudapestMetroDisplay.config import settings

logger = logging.getLogger(__name__)
brightness: float = 1.0

client: APIClient


def on_state_change(state: EntityState) -> None:
    """Process entity state changes from ESPHome."""
    global brightness

    # Check if this is a light entity
    if isinstance(state, LightState):
        brightness = state.brightness if hasattr(state, "brightness") else 1.0
        logger.debug(f"ESPHome Light brightness updated to {brightness * 100:.0f}%")


async def on_connect() -> None:
    """Perform actions after a successful connect."""
    try:
        # Subscribe to the state changes
        client.subscribe_states(on_state_change)
    except APIConnectionError:
        logger.exception(
            f"Error getting initial data for {settings.esphome.device_ip}",
        )
        # Re-connection logic will trigger after this
        await client.disconnect()


async def on_disconnect(expected_disconnect) -> None:  # noqa: ANN001
    """Run disconnect stuff on API disconnect."""
    logger.info(f"Disconnected changed to '{expected_disconnect}'")


async def on_connect_error(err: Exception) -> None:
    """Show connection errors."""
    logger.error(f"Failed to connect with error '{err}'")


async def connect_and_subscribe() -> None:
    """Connect to the ESPHome device and subscribe to state changes."""
    global client

    client = APIClient(
        str(settings.esphome.device_ip),
        6053,
        None,
        noise_psk=settings.esphome.api_key,
        client_info=f"BudapestMetroDisplay {__version__}",
    )

    # Build reconnect logic
    reconnect_logic = ReconnectLogic(
        client=client,
        on_connect=on_connect,
        on_disconnect=on_disconnect,
        zeroconf_instance=None,
        name="ESPHome",
        on_connect_error=on_connect_error,
    )

    await reconnect_logic.start()


def esphome_background_process(loop: AbstractEventLoop) -> None:
    """Run the asyncio loop."""
    asyncio.set_event_loop(loop)
    loop.run_forever()


def start_background_loop() -> AbstractEventLoop:
    """Start a background thread for asyncio for communicating with the device."""
    loop = asyncio.new_event_loop()
    threading.Thread(
        target=esphome_background_process,
        args=(loop,),
        daemon=True,
        name="ESPHome thread",
    ).start()
    return loop



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

# __init.py__

"""Background service for a BudapestMetroDisplay hardware LED display."""



#  MIT License
#
#  Copyright (c) 2024 denes44
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

# ruff: noqa: D103, S101, ANN001, ANN202

import os
import threading
from unittest.mock import MagicMock, patch

import pytest

# Mock environment variables before importing the module
os.environ["BKK_API_KEY"] = "test_api_key"

from BudapestMetroDisplay import (
    led_control,
    stops,
)
from BudapestMetroDisplay.main import settings


@pytest.fixture(autouse=True)
def mock_logger(monkeypatch) -> None:
    class MockLogger:
        def trace(self, message):
            pass

        def debug(self, message):
            pass

        def info(self, message):
            pass

        def warning(self, message):
            pass

        def error(self, message):
            pass

    mock_logger = MockLogger()
    monkeypatch.setattr("BudapestMetroDisplay.led_control.logger", mock_logger)


def test_find_key_by_value_returns_correct_key() -> None:
    d = {"a": 1, "b": 2, "c": 3}
    assert led_control.find_key_by_value(d, 2) == "b"


def test_find_key_by_value_returns_none_for_missing_value() -> None:
    d = {"a": 1, "b": 2, "c": 3}
    assert led_control.find_key_by_value(d, 4) is None


def test_find_keys_by_value_returns_correct_keys() -> None:
    d = {"a": 1, "b": 2, "c": 1}
    assert led_control.find_keys_by_value(d, 1) == ["a", "c"]


def test_find_keys_by_value_returns_empty_list_for_missing_value() -> None:
    d = {"a": 1, "b": 2, "c": 3}
    assert led_control.find_keys_by_value(d, 4) == []


def test_fade_color_fades_correctly(monkeypatch) -> None:
    led_index = 4
    current_color = (0, 0, 0)
    target_color = (255, 129, 63)
    steps = 10
    delay = 0.01

    def mock_update_sacn() -> None:
        pass

    monkeypatch.setattr(
        "BudapestMetroDisplay.led_control.update_sacn",
        mock_update_sacn,
    )
    led_control.fade_color(led_index, current_color, target_color, steps, delay)
    assert led_control.led_states[led_index * 3 : led_index * 3 + 3] == list(
        target_color,
    )


def test_reset_leds_to_default_sets_all_leds_correctly() -> None:
    # Change color of one LED to check if it is reset
    led_index = 4
    led_control.led_states[led_index * 3 : led_index * 3 + 3] = (255, 255, 255)

    # Reset all LEDs
    led_control.reset_leds_to_default()
    # Check if all LEDs are reset to default color
    assert led_control.led_states == [
        value for color in led_control.DEFAULT_COLORS for value in color
    ]


def test_reset_led_to_default_sets_specific_led_correctly(monkeypatch) -> None:
    # Change color of one LED to check if it is reset
    led_index = 4
    target_color = led_control.DEFAULT_COLORS[led_index]
    event = threading.Event()

    # Mock the fade_color function to set the target color directly
    def mock_fade_color(
        _led_index: int,
        _current_color: tuple[int, int, int],
        _target_color: tuple[int, int, int],
        _steps: int,
        _delay: float,
    ) -> None:
        led_control.led_states[_led_index * 3 : _led_index * 3 + 3] = _target_color
        event.set()

    monkeypatch.setattr("BudapestMetroDisplay.led_control.fade_color", mock_fade_color)

    # Change color of the LED
    led_control.led_states[led_index * 3 : led_index * 3 + 3] = (255, 255, 255)
    # Reset the LED
    led_control.reset_led_to_default(led_index)
    # Wait for the fade to finish
    event.wait(timeout=1)
    # Check if the LED is reset to default color
    assert led_control.led_states[led_index * 3 : led_index * 3 + 3] == list(
        target_color,
    )


def test_reset_led_to_default_sets_specific_led_correctly_with_no_alert(
    monkeypatch,
) -> None:
    # Change color of one LED to check if it is reset
    led_index = 6
    target_color = led_control.DEFAULT_COLORS[led_index]
    event = threading.Event()

    # Mock the fade_color function to set the target color directly
    def mock_fade_color(
        _led_index: int,
        _current_color: tuple[int, int, int],
        _target_color: tuple[int, int, int],
        _steps: int,
        _delay: float,
    ) -> None:
        led_control.led_states[_led_index * 3 : _led_index * 3 + 3] = _target_color
        event.set()

    monkeypatch.setattr("BudapestMetroDisplay.led_control.fade_color", mock_fade_color)

    # Change color of the LED
    led_control.led_states[led_index * 3 : led_index * 3 + 3] = (255, 255, 255)
    # Set alerts for the stopIds of this LED
    led_control.stops.stop_no_service["BKK_056215"] = False
    led_control.stops.stop_no_service["BKK_056216"] = True
    # Reset the LED
    led_control.reset_led_to_default(led_index)
    # Wait for the fade to finish
    event.wait(timeout=1)
    # Check if the LED is reset to default color
    assert led_control.led_states[led_index * 3 : led_index * 3 + 3] == list(
        target_color,
    )


def test_set_led_color_sets_color_correctly(monkeypatch) -> None:
    # Set the new color for the LED
    led_index = 4
    new_color = (0, 255, 0)
    event = threading.Event()

    # Mock the fade_color function to set the target color directly
    def mock_fade_color(
        _led_index: int,
        _current_color: tuple[int, int, int],
        _target_color: tuple[int, int, int],
        _steps: int,
        _delay: float,
    ) -> None:
        led_control.led_states[_led_index * 3 : _led_index * 3 + 3] = _target_color
        event.set()

    monkeypatch.setattr("BudapestMetroDisplay.led_control.fade_color", mock_fade_color)
    # Set the new color
    led_control.set_led_color(led_index, new_color)
    # Wait for the fade to finish
    event.wait(timeout=1)
    # Check if the LED is set to the new color
    assert led_control.led_states[led_index * 3 : led_index * 3 + 3] == list(new_color)


def test_set_led_color_combines_colors_correctly(monkeypatch) -> None:
    # Set the new color for the LED
    led_index = 4
    current_color = led_control.ROUTE_COLORS["BKK_5200"]
    new_color = led_control.ROUTE_COLORS["BKK_5300"]
    expected_color = (255, 0, 255)
    event = threading.Event()

    # Mock the fade_color function to set the target color directly
    def mock_fade_color(
        _led_index: int,
        _current_color: tuple[int, int, int],
        _target_color: tuple[int, int, int],
        _steps: int,
        _delay: float,
    ) -> None:
        led_control.led_states[_led_index * 3 : _led_index * 3 + 3] = _target_color
        event.set()

    monkeypatch.setattr("BudapestMetroDisplay.led_control.fade_color", mock_fade_color)
    # Set the LED color
    led_control.led_states[led_index * 3 : led_index * 3 + 3] = current_color
    # Set the LED to the new color
    led_control.set_led_color(led_index, new_color)
    # Wait for the fade to finish
    event.wait(timeout=1)
    # Check if the LED is set to the new color combined with the current color
    assert led_control.led_states[led_index * 3 : led_index * 3 + 3] == list(
        expected_color,
    )


def test_set_led_color_does_not_combine_colors(monkeypatch) -> None:
    # Set the new color for the LED
    led_index = 4
    current_color = led_control.ROUTE_COLORS_DIM["BKK_5200"]
    new_color = led_control.ROUTE_COLORS["BKK_5300"]
    expected_color = led_control.ROUTE_COLORS["BKK_5300"]
    event = threading.Event()

    # Mock the fade_color function to set the target color directly
    def mock_fade_color(
        _led_index: int,
        _current_color: tuple[int, int, int],
        _target_color: tuple[int, int, int],
        _steps: int,
        _delay: float,
    ) -> None:
        led_control.led_states[_led_index * 3 : _led_index * 3 + 3] = _target_color
        event.set()

    monkeypatch.setattr("BudapestMetroDisplay.led_control.fade_color", mock_fade_color)
    # Set the LED color
    led_control.led_states[led_index * 3 : led_index * 3 + 3] = current_color
    # Set the LED to the new color
    led_control.set_led_color(led_index, new_color)
    # Wait for the fade to finish
    event.wait(timeout=1)
    # Check if the LED is set to the new color combined with the current color
    assert led_control.led_states[led_index * 3 : led_index * 3 + 3] == list(
        expected_color,
    )


def test_get_led_color_returns_correct_color() -> None:
    led_index = 4
    expected_color = (100, 100, 100)
    led_control.led_states[led_index * 3 : led_index * 3 + 3] = expected_color
    assert led_control.get_led_color(led_index) == expected_color


def test_get_led_color_returns_none_for_invalid_index() -> None:
    assert led_control.get_led_color(-1) is None
    assert led_control.get_led_color(led_control.NUM_LEDS) is None


def test_activate_sacn() -> None:
    # Mock the sACNsender
    mock_sender = MagicMock()

    # Patch the sACNsender with the mock
    with patch(
        "BudapestMetroDisplay.led_control.sacn.sACNsender",
        return_value=mock_sender,
    ):
        # Call the function to test
        led_control.activate_sacn()

        # Verify that the sACNsender was started
        mock_sender.start.assert_called_once()
        # Verify that the output was activated
        mock_sender.activate_output.assert_called_once_with(
            led_control.settings.sacn.universe,
        )
        # Verify that the multicast setting was applied
        mock_sender.__getitem__.return_value.multicast = (
            led_control.settings.sacn.multicast
        )
        if not led_control.settings.sacn.multicast:
            # Verify that the destination was set if multicast is enabled
            mock_sender.__getitem__.return_value.destination = (
                led_control.settings.sacn.unicast_ip
            )


def test_deactivate_sacn() -> None:
    # Mock the sender
    mock_sender = MagicMock()

    # Patch the sender with the mock
    with patch("BudapestMetroDisplay.led_control.sender", mock_sender):
        # Call the function to test
        led_control.deactivate_sacn()

        # Verify that the stop method was called
        mock_sender.stop.assert_called_once()


def test_update_sacn() -> None:
    # Mock the sACNsender and its universe
    mock_sender = MagicMock()
    mock_universe = MagicMock()
    mock_sender.__getitem__.return_value = mock_universe

    # Patch the sender with the mock
    # Patch the sender with the mock
    with (
        patch("BudapestMetroDisplay.led_control.sender", mock_sender),
        patch.dict(
            "BudapestMetroDisplay.led_control.settings.esphome.__dict__",
            {"used": False},
        ),
    ):
        # Set the led_states to a known value
        led_control.led_states[:] = [10, 20, 30] * (len(led_control.led_states) // 3)
        expected_dmx_data = tuple(led_control.led_states)

        # Call the function to test
        led_control.update_sacn()

        # Verify that the dmx_data was updated correctly
        mock_universe.dmx_data = expected_dmx_data
        mock_sender.__getitem__.assert_called_with(
            led_control.settings.sacn.universe,
        )
        assert mock_universe.dmx_data == expected_dmx_data


def test_update_sacn_esphome(monkeypatch) -> None:  # noqa: ARG001
    # Mock the sACNsender and its universe
    mock_sender = MagicMock()
    mock_universe = MagicMock()
    mock_sender.__getitem__.return_value = mock_universe

    # Patch the sender with the mock
    with (
        patch("BudapestMetroDisplay.led_control.sender", mock_sender),
        patch.dict(
            "BudapestMetroDisplay.led_control.settings.esphome.__dict__",
            {"used": True},
        ),
        patch("BudapestMetroDisplay.esphome.brightness", 0.66),
    ):
        # Set the led_states to a known value
        led_control.led_states = [0, 0, 0] * led_control.NUM_LEDS
        # Set the first 3 LEDs to the expected result
        led_control.led_states[0:3] = [169, 85, 43]
        expected_dmx_data = tuple(led_control.led_states)
        # Set the first 3 LEDs to the starting value
        led_control.led_states[0:3] = [255, 128, 64]

        # Call the function to test
        led_control.update_sacn()

        # Verify that the dmx_data was updated correctly
        mock_universe.dmx_data = expected_dmx_data
        mock_sender.__getitem__.assert_called_with(
            led_control.settings.sacn.universe,
        )
        assert mock_universe.dmx_data == expected_dmx_data


def test_calculate_default_color_led12() -> None:
    led_test = 12  # Kálvin tér
    # Initialize the route status dictionary
    stops.stop_no_service = dict.fromkeys(stops.stops_led, False)
    led_control.calculate_default_color(led_test)
    assert led_control.DEFAULT_COLORS[led_test] == (
        0,
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for one direction of Kálvin tér M4
    stops.stop_no_service["BKK_056227"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        0,
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for the other direction of Kálvin tér M4
    stops.stop_no_service["BKK_056228"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5300"]
    )
    # Set the NO_SERVICE status for one direction of Kálvin tér M3
    stops.stop_no_service["BKK_F01290"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5300"]
    )
    # Set the NO_SERVICE status for one direction of Kálvin tér M3
    stops.stop_no_service["BKK_F01289"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (0, 0, 0)
    # Clear the NO_SERVICE status for one direction of Kálvin tér M4
    stops.stop_no_service["BKK_056227"] = False
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5400"]
    )


def test_calculate_default_color_led17() -> None:
    led_test = 17
    # Initialize the route status dictionary
    stops.stop_no_service = dict.fromkeys(stops.stops_led, False)
    led_control.calculate_default_color(led_test)
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        0,
        int(128 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for one direction of Batthyányi tér M2
    stops.stop_no_service["BKK_F00063"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        0,
        int(128 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for second direction of Batthyányi tér M2
    stops.stop_no_service["BKK_F00062"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_H5"]
    )
    # Set the NO_SERVICE status for one direction of Batthyányi tér H5
    stops.stop_no_service["BKK_09001187"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_H5"]
    )
    # Set the NO_SERVICE status for second direction of Batthyányi tér H5
    stops.stop_no_service["BKK_09001188"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_H5"]
    )
    # Set the NO_SERVICE status for third direction of Batthyányi tér H5
    stops.stop_no_service["BKK_09001189"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (0, 0, 0)
    # Clear the NO_SERVICE status for second direction of Batthyányi tér M2
    stops.stop_no_service["BKK_F00062"] = False
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5200"]
    )


def test_calculate_default_color_led22() -> None:
    led_test = 22  # Keleti pályaudvar
    # Initialize the route status dictionary
    stops.stop_no_service = dict.fromkeys(stops.stops_led, False)
    led_control.calculate_default_color(led_test)
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        int(128 * settings.led.dim_ratio),
        0,
    )
    # Set the NO_SERVICE status for one direction of Keleti pályaudvar M4
    stops.stop_no_service["BKK_056233"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        int(128 * settings.led.dim_ratio),
        0,
    )
    # Set the NO_SERVICE status for second direction of Keleti pályaudvar M4
    stops.stop_no_service["BKK_056234"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5200"]
    )
    # Set the NO_SERVICE status for one direction of Keleti pályaudvar M3
    stops.stop_no_service["BKK_F01336"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5200"]
    )
    # Set the NO_SERVICE status for second direction of Keleti pályaudvar M3
    stops.stop_no_service["BKK_F01335"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (0, 0, 0)
    # Clear the NO_SERVICE status for second direction of Keleti pályaudvar M4
    stops.stop_no_service["BKK_056234"] = False
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5400"]
    )


def test_calculate_default_color_led25() -> None:
    led_test = 25  # Örs vezér tere
    # Initialize the route status dictionary
    stops.stop_no_service = dict.fromkeys(stops.stops_led, False)
    led_control.calculate_default_color(led_test)
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        0,
        int(48 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for one direction of Örs vezér tere M2
    stops.stop_no_service["BKK_F01749"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_H8"]
    )
    # Set the NO_SERVICE status for one direction of Örs vezér tere H8
    stops.stop_no_service["BKK_19795278"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_H8"]
    )
    # Set the NO_SERVICE status for second direction of Örs vezér tere H8
    stops.stop_no_service["BKK_19795279"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_H8"]
    )
    # Set the NO_SERVICE status for third direction of Örs vezér tere H8
    stops.stop_no_service["BKK_19795280"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (0, 0, 0)
    # Clear the NO_SERVICE status for one direction of Örs vezér tere M2
    stops.stop_no_service["BKK_F01749"] = False
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert (
        led_control.DEFAULT_COLORS[led_test] == led_control.ROUTE_COLORS_DIM["BKK_5200"]
    )


def test_calculate_default_color_led19() -> None:
    led_test = 19  # Deák Ferenc tér
    # Initialize the route status dictionary
    stops.stop_no_service = dict.fromkeys(stops.stops_led, False)
    led_control.calculate_default_color(led_test)
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for one direction of Deák Ferenc tér M1
    stops.stop_no_service["BKK_F00963"] = True
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for second direction of Deák Ferenc tér M1
    stops.stop_no_service["BKK_F00962"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        0,
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for one direction of Örs vezér tere M2
    stops.stop_no_service["BKK_F00961"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        int(255 * settings.led.dim_ratio),
        0,
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for second direction of Örs vezér tere M2
    stops.stop_no_service["BKK_F00960"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        0,
        0,
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for one direction of Örs vezér tere M3
    stops.stop_no_service["BKK_F00955"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (
        0,
        0,
        int(255 * settings.led.dim_ratio),
    )
    # Set the NO_SERVICE status for second direction of Örs vezér tere M2
    stops.stop_no_service["BKK_F00954"] = True
    led_control.calculate_default_color(led_test)
    led_control.reset_leds_to_default()
    assert led_control.DEFAULT_COLORS[led_test] == (0, 0, 0)



#  MIT License
#
#  Copyright (c) 2024 denes44
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

# ruff: noqa: D101, D102, D103, D107, S101, ANN204, ANN201, ANN001

from datetime import datetime

from BudapestMetroDisplay.aps_helpers import (
    calculate_average_time_between_jobs,
    count_jobs_by_argument,
    find_soonest_job_by_argument,
    get_jobs_by_argument,
)


class MockScheduler:
    def __init__(self):
        self.jobs = []

    def add_job(self, job):
        self.jobs.append(job)

    def get_jobs(self):
        return self.jobs


class MockJob:
    def __init__(self, args=None, next_run_time=None):
        self.args = args or []
        self.next_run_time = next_run_time


def test_count_jobs_by_argument_counts_correctly() -> None:
    scheduler = MockScheduler()
    scheduler.add_job(MockJob(args=[1, 2, 3]))
    scheduler.add_job(MockJob(args=[1, 2, 3]))
    scheduler.add_job(MockJob(args=[4, 5, 6]))
    assert count_jobs_by_argument(scheduler, 2, 1) == 2


def test_count_jobs_by_argument_no_match() -> None:
    scheduler = MockScheduler()
    scheduler.add_job(MockJob(args=[1, 2, 3]))
    scheduler.add_job(MockJob(args=[4, 5, 6]))
    assert count_jobs_by_argument(scheduler, 7, 1) == 0


def test_get_jobs_by_argument_returns_correct_jobs() -> None:
    scheduler = MockScheduler()
    job1 = MockJob(args=[1, 2, 3])
    job2 = MockJob(args=[1, 2, 3])
    job3 = MockJob(args=[4, 5, 6])
    scheduler.add_job(job1)
    scheduler.add_job(job2)
    scheduler.add_job(job3)
    result = get_jobs_by_argument(scheduler, 2, 1)
    assert result == [job1, job2]


def test_get_jobs_by_argument_no_match() -> None:
    scheduler = MockScheduler()
    scheduler.add_job(MockJob(args=[1, 2, 3]))
    scheduler.add_job(MockJob(args=[4, 5, 6]))
    result = get_jobs_by_argument(scheduler, 7, 1)
    assert result == []


def test_find_soonest_job_by_argument_finds_correct_job() -> None:
    scheduler = MockScheduler()
    job1 = MockJob(args=[1, 2, 3], next_run_time=datetime(2024, 1, 1, 12, 0))
    job2 = MockJob(args=[1, 2, 3], next_run_time=datetime(2024, 1, 1, 10, 0))
    job3 = MockJob(args=[4, 5, 6], next_run_time=datetime(2024, 1, 1, 11, 0))
    scheduler.add_job(job1)
    scheduler.add_job(job2)
    scheduler.add_job(job3)
    result = find_soonest_job_by_argument(scheduler, 2, 1)
    assert result == job2


def test_find_soonest_job_by_argument_no_match() -> None:
    scheduler = MockScheduler()
    scheduler.add_job(
        MockJob(args=[1, 2, 3], next_run_time=datetime(2024, 1, 1, 12, 0)),
    )
    scheduler.add_job(
        MockJob(args=[4, 5, 6], next_run_time=datetime(2024, 1, 1, 10, 0)),
    )
    result = find_soonest_job_by_argument(scheduler, 7, 1)
    assert result is None


def calculate_average_time_between_jobs_calculates_correctly() -> None:
    job1 = MockJob(next_run_time=datetime(2024, 1, 1, 10, 0))
    job2 = MockJob(next_run_time=datetime(2024, 1, 1, 11, 0))
    job3 = MockJob(next_run_time=datetime(2024, 1, 1, 12, 0))
    result = calculate_average_time_between_jobs([job1, job2, job3])
    assert result == 3600.0


def calculate_average_time_between_jobs_not_enough_jobs() -> None:
    job1 = MockJob(next_run_time=datetime(2024, 1, 1, 10, 0))
    result = calculate_average_time_between_jobs([job1])
    assert result is None



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import logging
import sys
from pathlib import Path
from typing import Any

from pydantic import (
    DirectoryPath,
    Field,
    IPvAnyAddress,
    ValidationError,
    field_validator,
    model_validator,
)
from pydantic_core.core_schema import ValidationInfo
from pydantic_settings import BaseSettings, SettingsConfigDict

logger = logging.getLogger(__name__)


class LEDConfig(BaseSettings):
    """Class to store LED related settings."""

    dim_ratio: float = Field(
        default=0.25,
        ge=0,
        le=1,
        description="Dim ratio for the background brightness of the LEDs \
            (0 means totally off)",
    )
    fade_time: float = Field(
        default=0.5,
        ge=0,
        description="Fade time in seconds for the LED turn on and off action",
    )

    model_config = SettingsConfigDict(env_prefix="LED_", frozen=True)


class SACNConfig(BaseSettings):
    """Class to store sACN related settings."""

    multicast: bool = Field(
        default=True,
        description="Whether the sACN protocol should use multicast or unicast",
    )
    unicast_ip: IPvAnyAddress | None = Field(
        default=None,
        description="The destination IP address for unicast sACN",
    )
    universe: int = Field(
        default=1,
        ge=1,
        lt=64000,
        description="DMX universe to send out data with the sACN protocol",
    )
    fps: int = Field(default=60, ge=1, description="FPS limit")

    @field_validator("unicast_ip")
    @classmethod
    def check_unicast_ip(
        cls,
        value: IPvAnyAddress,
        info: ValidationInfo,
    ) -> IPvAnyAddress:
        """Check the validity of unicast_ip if multicast is turned off."""
        if "multicast" in info.data and not info.data["multicast"] and value is None:
            msg = "Must be a valid IP address when using unicast"
            raise ValueError(msg)
        return value

    model_config = SettingsConfigDict(env_prefix="SACN_", frozen=True)


class BKKConfig(BaseSettings):
    """Class to store BKK API related settings."""

    api_key: str = Field(
        pattern=(
            r"^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-"
            r"[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-"
            r"[a-fA-F0-9]{12}$"
        ),
        description="API key for the BKK OpenData portal",
    )
    api_update_interval: int = Field(
        default=2,
        gt=0,
        description="Delay between consecutive API calls in seconds",
    )
    api_update_realtime: int = Field(
        default=60,
        gt=0,
        description="Update frequency for realtime data in seconds",
    )
    api_update_regular: int = Field(
        default=1800,
        gt=0,
        description="Update frequency for regular data in seconds",
    )
    api_update_alerts: int = Field(
        default=600,
        gt=0,
        description="Update frequency for alerts for non-realtime routes in seconds",
    )

    model_config = SettingsConfigDict(env_prefix="BKK_", frozen=True)


class ESPHomeConfig(BaseSettings):
    """Class to store ESPHome related settings."""

    used: bool = Field(
        default=False,
        description="Whether to use brightness data from ESPHome \
            to determine the minimum brightness",
    )
    device_ip: IPvAnyAddress | None = Field(
        default=None,
        description="The IP address of the ESPHome device",
    )
    api_key: str | None = Field(
        default=None,
        pattern=r"^[A-Za-z0-9+/]{43}=$",
        description="The API key of the ESPHome device",
    )

    @field_validator("device_ip")
    @classmethod
    def check_unicast_ip(
        cls,
        value: IPvAnyAddress,
        info: ValidationInfo,
    ) -> IPvAnyAddress:
        """Check the validity of unicast_ip if ESPHome is used."""
        if "used" in info.data and info.data["used"] and value is None:
            msg = "Device IP must be filled out when using ESPHome"
            raise ValueError(msg)
        return value

    model_config = SettingsConfigDict(env_prefix="ESPHOME_", frozen=True)


class LogConfig(BaseSettings):
    """Class to store log related settings."""

    path: DirectoryPath = Field(
        default=Path("./log"),
        description="The directory to store log files",
    )

    model_config = SettingsConfigDict(env_prefix="LOG_", frozen=True)

    # Use `model_validator` to create the log directory if it doesn't exist
    @model_validator(mode="before")
    @classmethod
    def ensure_log_directory_exists(cls, values: Any) -> Any:
        """Check if the specified directory exists, and create if not."""
        path = values.get("path", Path("./log"))
        path = Path(path)  # Ensure `path` is a Path object

        if not path.exists():
            try:
                path.mkdir(parents=True, exist_ok=True)
            except BaseException as e:
                # TODO more specific exception handling
                msg = f"Unable to create log directory at {path}"
                raise ValueError(msg) from e

        # Update `path` back in `values`
        values["path"] = path
        return values


class AppConfig(BaseSettings):
    """Class to store application settings."""

    try:
        led: LEDConfig = LEDConfig()
        sacn: SACNConfig = SACNConfig()
        bkk: BKKConfig = BKKConfig()  # type: ignore[call-arg]
        esphome: ESPHomeConfig = ESPHomeConfig()
        log: LogConfig = LogConfig()
    except ValidationError:
        logger.exception("Configuration Error: Please check your environment variables")
        sys.exit(1)  # Exit the application with a non-zero status code

    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        frozen=True,
    )


settings = AppConfig()



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import logging
from datetime import datetime, time, timedelta
from random import randint
from typing import Any

import requests
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.background import BackgroundScheduler

from BudapestMetroDisplay import aps_helpers, led_control
from BudapestMetroDisplay._version import __version__
from BudapestMetroDisplay.config import settings
from BudapestMetroDisplay.stops import stop_no_service, stops_led

logger = logging.getLogger(__name__)
# Set the logging level for urllib3 to INFO
logging.getLogger("urllib3").setLevel(logging.INFO)
# Set logging level for requests to INFO
logging.getLogger("requests").setLevel(logging.INFO)

# Define base URL
API_BASE_URL: str = "https://futar.bkk.hu/api/query/v1/ws/otp/api/where/"

# Define the API request parameters for the different schedule updates
API_SCHEDULE_PARAMETERS: dict[str, dict[str, Any]] = {
    "REGULAR": {
        "minutesBefore": 0,
        "minutesAfter": round((settings.bkk.api_update_regular + 300) / 60),
        "limit": 200,
        "nextSchedule": timedelta(seconds=settings.bkk.api_update_regular),
    },
    "REALTIME": {
        "minutesBefore": 0,
        "minutesAfter": round(settings.bkk.api_update_realtime * 2 / 60),
        "limit": 100,
        "nextSchedule": timedelta(seconds=settings.bkk.api_update_realtime),
    },
}

# Default delay to turn off the LED
# when there is no separate arrival and departure time is available
ACTION_DELAY: dict[str, int] = {
    "BKK_5100": 10,  # M1
    "BKK_5200": 10,  # M2
    "BKK_5300": 10,  # M3
    "BKK_5400": 10,  # M4
    "BKK_H5": 15,  # H5
    "BKK_H6": 15,  # H6
    "BKK_H7": 15,  # H7
    "BKK_H8": 15,  # H8
    "BKK_H9": 15,  # H9
}

# Set minimum log level for APScheduler
logging.getLogger("apscheduler.executors.default").setLevel(logging.WARNING)
logging.getLogger("apscheduler.scheduler").setLevel(logging.WARNING)

# Initialize APScheduler with a persistent job store
# for scheduling the departures (when we turn on the LEDs)
departure_scheduler: BackgroundScheduler = BackgroundScheduler(
    jobstores={"default": MemoryJobStore()},  # Use in-memory job storage
)
departure_scheduler.start()

# Initialize scheduler with MemoryJobStore
# for scheduling turning off the LEDs after departure
led_scheduler: BackgroundScheduler = BackgroundScheduler(
    jobstores={"default": MemoryJobStore()},  # Use in-memory job storage
)
led_scheduler.start()

# Initialize scheduler with MemoryJobStore for scheduling the API updates
api_update_scheduler: BackgroundScheduler = BackgroundScheduler(
    jobstores={"default": MemoryJobStore()},  # Use in-memory job storage
)
api_update_scheduler.start()


def create_schedule_updates(
    stop_sets: tuple[tuple[str, tuple[str, ...]], ...],
    schedule_type: str,
) -> None:
    """Create jobs to update the data for the provided stops.

    The method puts the jobs in APScheduler which later will make the API calls.

    :param stop_sets: A tuple which consist the stop set name as a string
    and a tuple with the stop ids
    :param schedule_type: REGULAR or REALTIME,
    affects the API update parameters
    """
    # Store the current time when we started the update process
    start_time = datetime.now()

    # Check if the supplied schedule_type parameter is valid,
    # and the API call parameters are available in API_SCHEDULE_PARAMETERS
    if schedule_type not in API_SCHEDULE_PARAMETERS:
        logger.error(f"Invalid schedule type request: {schedule_type}")
        return

    logger.info(
        f"Starting updating the {schedule_type} schedules for stop set "
        f"({', '.join(stop_set for stop_set, _ in stop_sets)})",
    )

    for i, stop_set in enumerate(stop_sets):
        # Schedule the API calls from each other by settings.bkk.api_update_interval
        # starting from the current time
        delay: int = i * settings.bkk.api_update_interval
        job_time: datetime = start_time + timedelta(seconds=delay)  # job start time
        job_id: str = f"{stop_set[0]}_{schedule_type}"  # job reference id

        # Add the job to the scheduler
        api_update_scheduler.add_job(
            fetch_schedule_for_stops,
            "date",
            run_date=job_time,
            args=[stop_set, schedule_type],
            id=job_id,
            replace_existing=True,
            # If the job exists, it will be replaced with the new time
        )

        logger.debug(
            f"Scheduling {schedule_type} API updates for stop set "
            f"{stop_set[0]} at {job_time!s}.",
        )


def create_alert_updates(routes: tuple[str, ...]) -> None:
    """Create jobs in APScheduler to update the alerts for the provided routes.

    The method puts the jobs in APScheduler which later will make the API calls.

    :param routes: A tuple which consist the route ids
    """
    # Store the current time when we started the update process
    start_time = datetime.now()

    logger.info(f"Starting updating the alerts for routes {routes}")

    for i, route_id in enumerate(routes):
        # Schedule the API calls from each other by settings.bkk.api_update_interval
        # starting from the current time
        delay: int = i * settings.bkk.api_update_interval
        job_time: datetime = start_time + timedelta(seconds=delay)  # job start time
        job_id: str = f"{route_id}_ALERTS"  # job reference id

        # Add the job to the scheduler
        api_update_scheduler.add_job(
            fetch_alerts_for_route,
            "date",
            run_date=job_time,
            args=[route_id],
            id=job_id,
            replace_existing=True,
            # If the job exists, it will be replaced with the new time
        )

        logger.debug(
            f"Scheduling alerts API updates for routes {routes} at {job_time!s}.",
        )


def fetch_schedule_for_stops(
    stop_set: tuple[str, tuple[str, ...]],
    schedule_type: str,
) -> None:
    """Send API request to fetch the schedule for the selected stops.

    Callback function for the APScheduler jobs.
    Makes an arrivals-and-departures-for-stop API request to the BKK OpenData server
    for the selected stops.

    :param stop_set: A tuple which consist the stop set name as a string
        and a tuple with the stop ids we want to get the schedules for
    :param schedule_type: REGULAR or REALTIME, affects the API update parameters
    """
    # Calculate next schedule time
    if schedule_type == "REALTIME" and (
        time(0, 30) <= datetime.now().time() <= time(4, 0)
    ):
        # When the current time is between 00:30 and 04:00
        # schedule the next REALTIME update for 04:00
        job_time = datetime.combine(datetime.today(), time(4, 0))
    else:
        # Otherwise schedule the next update according to the configuration
        job_time = (
            datetime.now() + API_SCHEDULE_PARAMETERS[schedule_type]["nextSchedule"]
        )

    url: str = f"{API_BASE_URL}arrivals-and-departures-for-stop"

    headers: dict[str, str] = {"Accept": "application/json"}

    # Get schedule data for all stops in the stop set
    params: dict[str, str | int | list[str]] = {
        "stopId": list(stop_set[1]),
        "minutesBefore": API_SCHEDULE_PARAMETERS[schedule_type]["minutesBefore"],
        "minutesAfter": API_SCHEDULE_PARAMETERS[schedule_type]["minutesAfter"],
        "limit": API_SCHEDULE_PARAMETERS[schedule_type]["limit"],
        "onlyDepartures": "false",
        "appVersion": f"BudapestMetroDisplay {__version__}",
        "version": "4",
        "includeReferences": "routes,alerts",
        "key": settings.bkk.api_key,
    }
    try:
        response = requests.get(url, headers=headers, params=params, timeout=5)

        if response.status_code == 200:
            # Recalculate schedule intervals for REGULAR updates
            if schedule_type == "REGULAR":
                calculate_schedule_interval((response.json()), stop_set[0])

            latest_departure_time: int = store_departures(response.json(), stop_set[0])

            if schedule_type != "REALTIME" and latest_departure_time == -1:
                job_time = datetime.now() + timedelta(minutes=1)
                logger.debug(
                    f"There were no departures during {schedule_type} schedule update "
                    f"for stop set {stop_set[0]}. "
                    f"Next update scheduled for {job_time!s}",
                )
            elif (
                schedule_type != "REALTIME"
                and datetime.fromtimestamp(latest_departure_time) < job_time
            ):
                job_time = datetime.fromtimestamp(latest_departure_time) - timedelta(
                    minutes=5,
                )
                logger.debug(
                    f"The calculated next {schedule_type} schedule update "
                    f"for stop set {stop_set[0]} is later than the last departure "
                    f"we stored, so next update time was adjusted accordingly. "
                    f"Next update scheduled for {job_time!s}",
                )
            else:
                logger.debug(
                    f"Successfully updated {schedule_type} schedules for stop set "
                    f"{stop_set[0]}. Next update scheduled for {job_time!s}",
                )
        else:
            # Reschedule the failed action for 1 minute later
            job_time = datetime.now() + timedelta(minutes=1)

            logger.error(
                f"Failed to update {schedule_type} schedules for stop set "
                f"{stop_set[0]}: {response.status_code}. "
                f"Rescheduled for {job_time!s}.",
            )
    except requests.exceptions.JSONDecodeError as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            "The response did not contain valid JSON data when updating "
            f"{schedule_type} schedules for stop set "
            f"{stop_set[0]}. Rescheduled for {job_time!s}.",
        )
        logger.warning(e)
    except requests.exceptions.InvalidJSONError as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            "The response contained invalid JSON data when updating "
            f"{schedule_type} schedules for stop set "
            f"{stop_set[0]}. Rescheduled for {job_time!s}.",
        )
        logger.warning(e)
    except requests.exceptions.ReadTimeout as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            f"Timeout occurred when updating {schedule_type} schedules for stop set "
            f"{stop_set[0]}. Rescheduled for {job_time!s}.",
        )
        logger.warning(e)
    except requests.exceptions.ConnectionError as e:
        job_time = datetime.now() + timedelta(minutes=5)
        logger.warning(
            f"Connection error when updating {schedule_type} schedules for stop set "
            f"{stop_set[0]}. Rescheduled for {job_time!s}.",
        )
        logger.warning(e)
    except requests.exceptions.RequestException as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            f"Error when updating {schedule_type} schedules for stop set {stop_set[0]}."
            f"Rescheduled for {job_time!s}.",
        )
        logger.warning(e)

    job_id: str = f"{stop_set[0]}_{schedule_type}"

    api_update_scheduler.add_job(
        fetch_schedule_for_stops,
        "date",
        run_date=job_time,
        args=[stop_set, schedule_type],
        id=job_id,
        replace_existing=True,
        # If the job exists, it will be replaced with the new time
    )


def fetch_alerts_for_route(route_id: str) -> None:
    """Send API request to fetch the alerts for a selected route.

    Callback function for the APScheduler jobs.
    Makes a route-details API request to the BKK OpenData server
    for the supplied route_id.

    :param route_id: The id of the stop we want to get the details for
    """
    # Calculate next schedule time
    job_time = datetime.now() + timedelta(seconds=settings.bkk.api_update_alerts)

    url: str = f"{API_BASE_URL}route-details"

    headers: dict[str, str] = {"Accept": "application/json"}

    params: dict[str, str | int] = {
        "routeId": route_id,
        "appVersion": f"BudapestMetroDisplay {__version__}",
        "version": "4",
        "includeReferences": "alerts",
        "key": settings.bkk.api_key,
    }

    try:
        response = requests.get(url, headers=headers, params=params, timeout=5)

        if response.status_code == 200:
            process_alerts(response.json(), route_id)

            logger.debug(
                f"Successfully updated alerts for route {route_id}. "
                f"Next update scheduled for {job_time!s}",
            )
        else:
            # Reschedule the failed action for 1 minute later
            job_time = datetime.now() + timedelta(minutes=1)

            logger.error(
                f"Failed to update alerts for route {route_id}: {response.status_code}."
                f"Rescheduled for {job_time!s}.",
            )
    except requests.exceptions.JSONDecodeError as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            "The response did not contain valid JSON data when updating "
            f"alerts for route {route_id}. "
            f"Next update scheduled for {job_time!s}",
        )
        logger.warning(e)
    except requests.exceptions.InvalidJSONError as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            "The response contained invalid JSON data when updating "
            f"alerts for route {route_id}. "
            f"Next update scheduled for {job_time!s}",
        )
        logger.warning(e)
    except requests.exceptions.ReadTimeout as e:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.warning(
            f"Timeout occurred when updating alerts for route {route_id}. "
            f"Next update scheduled for {job_time!s}",
        )
        logger.warning(e)
    except requests.exceptions.ConnectionError:
        job_time = datetime.now() + timedelta(minutes=5)
        logger.exception(
            f"Connection error when updating alerts for route {route_id}. "
            f"Next update scheduled for {job_time!s}",
        )
    except requests.exceptions.RequestException:
        job_time = datetime.now() + timedelta(minutes=1)
        logger.exception(
            f"Error when updating alerts for route {route_id}. "
            f"Next update scheduled for {job_time!s}",
        )

    job_id: str = f"{route_id}_ALERTS"

    api_update_scheduler.add_job(
        fetch_alerts_for_route,
        "date",
        run_date=job_time,
        args=[route_id],
        id=job_id,
        replace_existing=True,
        # If the job exists, it will be replaced with the new time
    )


def store_departures(json_response: Any, reference_id: str) -> int:
    """Process the API response and store the departures.

    Processes the ArrivalsAndDeparturesForStopOTPMethodResponse API response
    from the arrivals-and-departures-for-stop method.

    As a result stores the retrieved departures in an APScheduler instance
    which will control the LED states.

    :param json_response: JSON return data from the BKK OpenData API
    :param reference_id: Internal reference to identify the API request in the logs
    :return: Timestamp of the latest departure in the data provided,
        -1 if there is no valid departures
    """
    # Check if JSON looks valid
    if (
        not json_response
        or "data" not in json_response
        or "entry" not in json_response["data"]
    ):
        logger.error(
            f"No valid schedule data in API response for stop(s) {reference_id}",
        )
        return -1

    # Check if we exceeded the query limit
    if json_response["data"].get("limitExceeded", "false") == "true":
        logger.warning(f"Query limit is exceeded when updating stop(s) {reference_id}")

    # Get routeId
    if (
        "routeIds" in json_response["data"]["entry"]
        and len(json_response["data"]["entry"]["routeIds"]) > 0
    ):
        route_id: str = json_response["data"]["entry"].get("routeIds", [])[0]
    elif (
        "routes" in json_response["data"]["references"]
        and len(json_response["data"]["references"]["routes"]) > 0
    ):
        route_id = next(json_response["data"]["references"]["routes"].keys())
    else:
        logger.warning(
            f"No route IDs found or the list is empty when updating stop(s) "
            f"{reference_id}",
        )
        return -1

    # Get stopId, when there is only one stop in the response
    stop_id_global: str = json_response["data"]["entry"].get("stopId", None)

    # Get stopTimes from TransitArrivalsAndDepartures
    stop_times = json_response["data"]["entry"].get("stopTimes", [])
    if len(stop_times) == 0:
        logger.debug(f"No schedule data found when updating stop(s) {reference_id}")

    latest_departure_time: int = -1
    # Iterate through the TransitScheduleStopTimes in the TransitArrivalsAndDepartures
    for stop_time in stop_times:
        trip_id: str = stop_time.get("tripId")

        # Get stopId, when the response contains schedules for multiple stops
        stop_id = stop_time.get("stopId") if "stopId" in stop_time else stop_id_global

        # Check if we are interested in the provided stopId
        if stop_id not in stops_led:
            logger.debug(
                f"Got update for stop {stop_id}, route {route_id}, "
                f"but we don't need that, skipping",
            )
            continue

        # Get the predicted or scheduled departure and arrival times
        # CASE #1: Both arrival and departure time available as predicted
        # [middle stop with realtime data]
        arrival_time: int
        delay: int
        if (
            "predictedArrivalTime" in stop_time
            and "predictedDepartureTime" in stop_time
            and not stop_time.get("uncertain", False)
        ):
            # Arrival time is different from the departure,
            # lets use the difference between them for the LED turn off delay
            if stop_time.get("predictedArrivalTime") != stop_time.get(
                "predictedDepartureTime",
            ):
                arrival_time = stop_time.get("predictedArrivalTime")
                delay = stop_time.get("predictedDepartureTime") - stop_time.get(
                    "predictedArrivalTime",
                )
            # Arrival time is the same as the departure,
            # use predefined delay for LED turn off
            else:
                arrival_time = (
                    stop_time.get("predictedDepartureTime") - ACTION_DELAY[route_id]
                )
                delay = ACTION_DELAY[route_id]
        # CASE #2: Only predicted arrival time is available
        # [end stop with realtime data]
        # Use the predefined delay for LED turn off
        elif "predictedArrivalTime" in stop_time and not stop_time.get(
            "uncertain",
            False,
        ):
            arrival_time = (
                stop_time.get("predictedArrivalTime") - ACTION_DELAY[route_id]
            )
            delay = ACTION_DELAY[route_id]
        # CASE #3: Only predicted departure time is available
        # [start stop with realtime data]
        # Use the predefined delay for LED turn off
        elif "predictedDepartureTime" in stop_time and not stop_time.get(
            "uncertain",
            False,
        ):
            arrival_time = (
                stop_time.get("predictedDepartureTime") - ACTION_DELAY[route_id]
            )
            delay = ACTION_DELAY[route_id]
        # CASE #4: Both arrival and departure time available as scheduled time
        # [middle stop, no realtime data]
        elif "arrivalTime" in stop_time and "departureTime" in stop_time:
            # Arrival time is different from the departure,
            # lets use the difference between them for the LED turn off delay
            if stop_time.get("arrivalTime") != stop_time.get("departureTime"):
                arrival_time = stop_time.get("arrivalTime") + randint(-3, 3)
                delay = stop_time.get("departureTime") - stop_time.get("arrivalTime")
            # Arrival time is the same as the departure,
            # use predefined delay for LED turn off
            else:
                arrival_time = (
                    stop_time.get("departureTime")
                    - ACTION_DELAY[route_id]
                    + randint(-3, 3)
                )
                delay = ACTION_DELAY[route_id]
        # CASE #5: Only scheduled arrival time is available
        # [end stop with no realtime data]
        # Use the predefined delay for LED turn off
        elif "arrivalTime" in stop_time:
            arrival_time = (
                stop_time.get("arrivalTime") - ACTION_DELAY[route_id] + randint(-3, 3)
            )
            delay = ACTION_DELAY[route_id]
        # CASE #6: Only scheduled departure time is available
        # [start stop with no realtime data]
        # Use the predefined delay for LED turn off
        elif "departureTime" in stop_time:
            arrival_time = (
                stop_time.get("departureTime") - ACTION_DELAY[route_id] + randint(-3, 3)
            )
            delay = ACTION_DELAY[route_id]
        # CASE #7: No valid time data is available
        else:
            logger.debug(
                f"No valid arrival/departure time found "
                f"when updating stop {stop_id}, route {route_id}",
            )
            continue

        # We processed valid schedule data for this stop,
        # that means that the stop is operational
        if stop_no_service[stop_id]:
            stop_no_service[stop_id] = False
            # Reset the LED to the default color
            led_control.calculate_default_color(stops_led[stop_id])
            # Change the LED color to the default color
            # if there is no ongoing LED action for this LED
            if not led_control.led_locks[stops_led[stop_id]].locked():
                led_control.reset_led_to_default(stops_led[stop_id])

        # Schedule the action 10 seconds before the departure."""
        job_id: str = f"{stop_id}+{trip_id}"
        job_time: datetime = datetime.fromtimestamp(arrival_time)

        latest_departure_time = max(latest_departure_time, arrival_time)

        if job_time > datetime.now():
            departure_scheduler.add_job(
                action_to_execute,
                "date",
                run_date=job_time,
                args=[stop_id, route_id, trip_id, job_time, delay],
                id=job_id,
                replace_existing=True,
                # If the job exists, it will be replaced with the new time
            )

            logger.trace(  # type: ignore[attr-defined]
                f"Scheduled action for departure: stop_id={stop_id}, "
                f"route_id={route_id}, trip_id={trip_id}, "
                f"departure_time={datetime.fromtimestamp(arrival_time)!s}",
            )
        else:
            logger.trace(  # type: ignore[attr-defined]
                f"Action for departure: stop_id={stop_id}, route_id={route_id}, "
                f"trip_id={trip_id}, "
                f"departure_time={datetime.fromtimestamp(arrival_time)!s} "
                f"was in the past, skipping",
            )

    # Check if there is any active alerts in the response and process them
    alerts = json_response["data"]["entry"].get("alertIds", [])
    if len(alerts) > 0:
        process_alerts(json_response, reference_id)

    return latest_departure_time


def action_to_execute(
    stop_id: str,
    route_id: str,
    trip_id: str,
    job_time: datetime,
    delay: int,
) -> None:
    """Execute actions for a departure.

    Callback function for a stop schedule.
    Changes the LED of the specified stop(stop_id) to the associated color,
    and schedules the turn-off of the LED in APScheduler after the supplied delay.

    :param stop_id: stopId from the BKK OpenData API
    :param route_id: routeId from the BKK OpenData API
    :param trip_id: tripId from the BKK OpenData API
    :param job_time: The time this job was scheduled in APScheduler
    :param delay: The amount of time needs to be elapsed in seconds
        between the turn on and turn off action
    """
    if job_time < datetime.now() - timedelta(seconds=20):
        logger.trace(  # type: ignore[attr-defined]
            "Action trigger time is in the past, skipping",
        )
        return

    logger.trace(  # type: ignore[attr-defined]
        f"Action triggered for stop: {stop_id}, route: {route_id}, trip: {trip_id}, "
        f"LED off delay: {delay} sec",
    )

    led_id: int = stops_led[stop_id]

    # Change LED color according to the color of the route
    led_control.set_led_color(led_id, led_control.ROUTE_COLORS[route_id])

    # Schedule an action at the departure time to turn the LED back to the default value
    led_job_time = job_time + timedelta(seconds=delay)

    led_scheduler.add_job(
        led_control.reset_led_to_default,
        "date",
        run_date=led_job_time,
        args=[led_id],
        id=str(led_id),
        replace_existing=True,
        # If the job exists, it will be replaced with the new time
    )


def calculate_schedule_interval(json_response: Any, reference_id: str) -> None:
    """Process API response to determine the interval between schedules for a route.

    Process the ArrivalsAndDeparturesForStopOTPMethodResponse API response
    from the arrivals-and-departures-for-stop method.
    It checks time between the schedules for each route
    and updates the ACTION_DELAY dictionary accordingly.

    :param json_response: JSON return data from the BKK OpenData API
    :param reference_id: Internal reference to identify the API request in the logs
    :return:
    """
    # Check if JSON looks valid
    if (
        not json_response
        or "data" not in json_response
        or "entry" not in json_response["data"]
    ):
        logger.error(
            f"No valid schedule data in API response "
            f"for schedule intervals for {reference_id}",
        )
        return

    # Check if we exceeded the query limit
    if json_response["data"].get("limitExceeded", "false") == "true":
        logger.warning(
            f"Query limit is exceeded when updating "
            f"schedule intervals for {reference_id}",
        )

    # Get routeId
    route_id: str
    if (
        "routeIds" in json_response["data"]["entry"]
        and len(json_response["data"]["entry"]["routeIds"]) > 0
    ):
        route_id = json_response["data"]["entry"].get("routeIds", [])[0]
    elif (
        "routes" in json_response["data"]["references"]
        and len(json_response["data"]["references"]["routes"]) > 0
    ):
        route_id = next(json_response["data"]["references"]["routes"].keys())
    else:
        logger.warning(
            f"No route IDs found or the list is empty when updating "
            f"schedule intervals for {reference_id}",
        )
        return

    # Get stopTimes from TransitArrivalsAndDepartures
    stop_times = json_response["data"]["entry"].get("stopTimes", [])
    if len(stop_times) < 2:
        logger.debug(
            f"Not enough schedule data found when updating "
            f"schedule intervals for {reference_id}",
        )
        return

    # Variable to store the necessary data from the first two relevant stops
    stop_id: str = stop_times[0].get("stopId", "")
    stop_headsign: str = stop_times[0].get("stopHeadsign", "")
    last_time: int = stop_times[0].get("departureTime", 0)

    deltas: list[int] = []

    # Iterate through the TransitScheduleStopTimes
    # in the TransitArrivalsAndDepartures
    for stop_time in stop_times[1:]:  # skip the first item
        # Check if the stopId and stopHeadsign is the same as the first one
        if (
            stop_time.get("stopId") != stop_id
            or stop_time.get("stopHeadsign") != stop_headsign
        ):
            continue

        current_time: int = stop_time.get("departureTime", 0)
        delta: int = current_time - last_time
        deltas.append(delta)
        last_time = current_time

    avg = sum(deltas) / len(deltas) / 60 if deltas else -1

    # Update the ACTION_DELAY dictionary according to the calculated difference
    if avg > 0:
        if route_id.startswith("BKK_5"):  # Subway
            if avg <= 2:
                ACTION_DELAY[route_id] = 15
            elif avg < 5.5:
                ACTION_DELAY[route_id] = 20
            else:
                ACTION_DELAY[route_id] = 30
        elif route_id.startswith("BKK_H"):  # Suburban railway
            if avg < 5.5:
                ACTION_DELAY[route_id] = 20
            elif avg < 10.5:
                ACTION_DELAY[route_id] = 30
            else:
                ACTION_DELAY[route_id] = 45
    # No valid schedule data found, set the default delay
    elif avg == -1:
        if route_id.startswith("BKK_5"):  # Subway
            ACTION_DELAY[route_id] = 30
        elif route_id.startswith("BKK_H"):  # Suburban railway
            ACTION_DELAY[route_id] = 45

    logger.debug(
        f"Recalculated LED turn off delay for route {route_id}, schedule interval:  "
        f"{avg:.1f} min, LED delay: {ACTION_DELAY[route_id]} sec",
    )


def process_alerts(json_response: Any, reference_id: str) -> None:
    """Process API response to determine the interval between schedules for a route.

    Process the ArrivalsAndDeparturesForStopOTPMethodResponse API response
    from the arrivals-and-departures-for-stop method.
    It checks time between the schedules for each route
    and updates the ACTION_DELAY dictionary accordingly.

    :param json_response: JSON return data from the BKK OpenData API
    :param reference_id: Internal reference to identify the API request in the logs
    :return:
    """
    # Check if JSON looks valid
    if (
        not json_response
        or "data" not in json_response
        or "references" not in json_response["data"]
        or "alerts" not in json_response["data"]["references"]
    ):
        logger.error(f"No valid alerts data in API response for stop(s) {reference_id}")
        return

    # Get stopTimes from TransitArrivalsAndDepartures
    alerts = json_response["data"]["references"].get("alerts")
    if len(alerts) == 0:
        logger.debug(f"No alert data found when updating stop(s) {reference_id}")

    # Iterate through the TransitScheduleStopTimes in the TransitArrivalsAndDepartures
    for alert_details in alerts.values():
        # If the start time of the alert is in the future, return False
        if alert_details["start"] > datetime.now().timestamp():
            continue

        # Iterate the TransitAlertRoutes in the TransitAlert
        for route in alert_details["routes"]:
            route_id: str = route.get("routeId", "")
            effect_type: str = route.get("effectType", "")

            # If the effectType is not NO_SERVICE
            # we don't process the TransitAlertRoute anymore
            if effect_type != "NO_SERVICE":
                continue

            # Iterate through stopIds in the TransitAlertRoute
            for stop_id in route.get("stopIds", []):
                # Check if we are interested in the stopId
                if stop_id not in stops_led:
                    continue

                # Check whether the stop is operational now
                if not stop_no_service[stop_id]:
                    # Check if we have a schedule for this stop_id,
                    # because if we have, then the stop is not really out of service
                    soonest_job = aps_helpers.find_soonest_job_by_argument(
                        departure_scheduler,
                        stop_id,
                        0,
                    )

                    if soonest_job is not None:
                        # We found at least one schedule for this stop,
                        # so we'll ignore the NO_SERVICE alert
                        logger.debug(
                            f"Found NO_SERVICE alert {alert_details['id']} "
                            f"for stop {stop_id}, route {route_id}, "
                            f"but there are active schedules for that stop, "
                            f"so we'll ignore that",
                        )
                    else:
                        # No schedule found for this stop,
                        # so we'll process the NO_SERVICE alert
                        logger.debug(
                            f"Found NO_SERVICE alert {alert_details['id']} "
                            f"for stop {stop_id}, route {route_id}",
                        )
                        # Set the no_service flag for this stop
                        stop_no_service[stop_id] = True

                        # Calculate the default color for this stop according to
                        # which route is operation for the stop
                        led_control.calculate_default_color(stops_led[stop_id])
                        # Change the LED color to the default color
                        led_control.reset_led_to_default(stops_led[stop_id], fade=False)



#  MIT License
#
#  Copyright (c) 2024 [fullname]
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom
#  the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

import inspect
import logging
import threading
import time
import uuid
from math import ceil
from threading import Lock

import sacn
from sacn import sACNsender

from BudapestMetroDisplay import stops
from BudapestMetroDisplay._version import __version__
from BudapestMetroDisplay.config import settings
from BudapestMetroDisplay.stops import common_stops

logger = logging.getLogger(__name__)

# The default LED color for the routes
ROUTE_COLORS: dict[str, tuple[int, int, int]] = {
    "BKK_5100": (255, 255, 0),  # M1
    "BKK_5200": (255, 0, 0),  # M2
    "BKK_5300": (0, 0, 255),  # M3
    "BKK_5400": (0, 255, 0),  # M4
    "BKK_H5": (255, 0, 255),  # H5
    "BKK_H6": (255, 0, 255),  # H6
    "BKK_H7": (255, 0, 255),  # H7
    "BKK_H8": (255, 0, 255),  # H8
    "BKK_H9": (255, 0, 255),  # H9
}

# The default dimmed LED color for the routes
ROUTE_COLORS_DIM: dict[str, tuple[int, int, int]] = {
    route: (
        int(color[0] * settings.led.dim_ratio),
        int(color[1] * settings.led.dim_ratio),
        int(color[2] * settings.led.dim_ratio),
    )
    for route, color in ROUTE_COLORS.items()
}
# Define default colors for each LED
DEFAULT_COLORS: list[tuple[int, int, int]] = [
    *([ROUTE_COLORS_DIM["BKK_H6"]] * 3),  # H6
    *([ROUTE_COLORS_DIM["BKK_H7"]] * 3),  # H7
    *([ROUTE_COLORS_DIM["BKK_5400"]] * 6),  # M4
    (
        0,
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
    ),  # M4/M3 Kálvin tér
    *([ROUTE_COLORS_DIM["BKK_5400"]] * 2),  # M4
    *([ROUTE_COLORS_DIM["BKK_5200"]] * 2),  # M2
    (
        int(255 * settings.led.dim_ratio),
        0,
        int(128 * settings.led.dim_ratio),
    ),  # M2/H5 Batthyány tér
    *([ROUTE_COLORS_DIM["BKK_5200"]] * 1),  # M2
    (
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
        int(255 * settings.led.dim_ratio),
    ),  # M1/M2/M3 Deák Ferenc tér
    *([ROUTE_COLORS_DIM["BKK_5200"]] * 2),  # M2
    (
        int(255 * settings.led.dim_ratio),
        int(128 * settings.led.dim_ratio),
        0,
    ),  # M2/M4 Keleti pályaudvar
    *([ROUTE_COLORS_DIM["BKK_5200"]] * 2),  # M2
    (
        int(255 * settings.led.dim_ratio),
        int(0 * settings.led.dim_ratio),
        int(48 * settings.led.dim_ratio),
    ),  # M2/H9 Örs vezér tere
    *([ROUTE_COLORS_DIM["BKK_H9"]] * 1),  # H9
    *([ROUTE_COLORS_DIM["BKK_H5"]] * 8),  # H5
    *([ROUTE_COLORS_DIM["BKK_5300"]] * 18),  # M3
    *([ROUTE_COLORS_DIM["BKK_5100"]] * 10),  # M1
]
DEFAULT_COLORS_ORIGINAL = list(DEFAULT_COLORS)

# Number of LEDs
NUM_LEDS: int = len(DEFAULT_COLORS)
# Initialize LED states with (0, 0, 0)
led_states: list[int] = [0, 0, 0] * NUM_LEDS
# Lock to synchronize access to LED states (led_states)
led_lock: Lock = threading.Lock()
# Initialize a dictionary of locks for each LED index
led_locks: dict[int, Lock] = {i: threading.Lock() for i in range(NUM_LEDS)}

# sACN sender interface
sender: sACNsender | None = None


def find_key_by_value(d: dict[str, int], target_value: int) -> str | None:
    """Find the supplied value (target_value) in the supplied dictionary.

    :param d: A dictionary to look for a value
    :param target_value: The value to look in the dictionary
    :return: Returns the key of the value if found, None otherwise
    """
    for key, value in d.items():
        if value == target_value:
            return key
    return None  # Return None if the value is not found


def find_keys_by_value(d: dict[str, int], target_value: int) -> list[str]:
    """Find all keys in the supplied dictionary that match the target_value.

    :param d: A dictionary to look for values
    :param target_value: The value to look for in the dictionary
    :return: A list of keys that match the value, or an empty list if none are found
    """
    return [key for key, value in d.items() if value == target_value]


def fade_color(
    led_index: int,
    current_color: tuple[int, int, int],
    target_color: tuple[int, int, int],
    steps: int = 100,
    delay: float = 0.05,
) -> None:
    """Fade from the current color to the target color of a specific LED.

    :param led_index: Index of the LED to update
    :param current_color: Tuple (r, g, b) representing the starting RGB color
    :param target_color: Tuple (r, g, b) representing the target RGB color
    :param steps: Number of steps in the fading process
    :param delay: Time to wait between each step in seconds
    """
    r1, g1, b1 = current_color
    r2, g2, b2 = target_color

    with led_locks[led_index]:  # Acquire the lock for the specific LED index
        for i in range(steps + 1):
            # Calculate the intermediate color
            r = int(r1 + (r2 - r1) * (i / steps))
            g = int(g1 + (g2 - g1) * (i / steps))
            b = int(b1 + (b2 - b1) * (i / steps))

            # Set the LED to the intermediate color
            with led_lock:
                led_states[led_index * 3 : led_index * 3 + 3] = [r, g, b]

            # Update the sACN values
            update_sacn()

            time.sleep(delay)

        # Ensure the final color is set correctly,
        # to avoid rounding errors dealing with floats
        with led_lock:
            led_states[led_index * 3 : led_index * 3 + 3] = [r2, g2, b2]
        update_sacn()


def calculate_default_color(led_index: int) -> None:
    """Recalculate the default LED colors.

    The method considers the active alerts for every stop,
    and calculates the default color of the stops according to the active routes.
    """
    # Check whether this LED belongs to multiple routes
    if led_index not in common_stops:
        # LED belongs to single route
        stop_ids: list[str] = [
            sid for sid, idx in stops.stops_led.items() if idx == led_index
        ]

        if all(stops.stop_no_service.get(sid, False) for sid in stop_ids):
            DEFAULT_COLORS[led_index] = (0, 0, 0)
        else:
            DEFAULT_COLORS[led_index] = DEFAULT_COLORS_ORIGINAL[led_index]
    else:
        # LED belongs to multiple routes
        route_status: dict[
            str,
            bool,
        ] = {}  # Status of each route for the specific stop/LED
        for route in common_stops[led_index]:
            route_id: str = route["route_id"]
            stop_ids = route["stop_ids"]

            # Check if all stops in the stop_ids list
            # have no service status for the current route
            if all(stops.stop_no_service.get(stop_id) for stop_id in stop_ids):
                route_status[route_id] = False
            else:
                route_status[route_id] = True

        # Determine the default color based on
        # which routes are operational at the specific stop/LED
        if led_index == 12:  # Kálvin tér: M3,M4
            if (
                route_status["BKK_5300"] and route_status["BKK_5400"]
            ):  # Both M3 and M4 are operational
                DEFAULT_COLORS[led_index] = (
                    0,
                    int(255 * settings.led.dim_ratio),
                    int(255 * settings.led.dim_ratio),
                )
            elif route_status["BKK_5300"]:  # Only M3 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_5300"]
            elif route_status["BKK_5400"]:  # Only M4 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_5400"]
            else:
                DEFAULT_COLORS[led_index] = (0, 0, 0)
        elif led_index == 17:  # Batthyány tér: M2,H5
            if (
                route_status["BKK_5200"] and route_status["BKK_H5"]
            ):  # Both M2 and H5 are operational
                DEFAULT_COLORS[led_index] = (
                    int(255 * settings.led.dim_ratio),
                    0,
                    int(128 * settings.led.dim_ratio),
                )
            elif route_status["BKK_5200"]:  # Only M2 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_5200"]
            elif route_status["BKK_H5"]:  # Only H5 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_H5"]
            else:
                DEFAULT_COLORS[led_index] = (0, 0, 0)
        elif led_index == 19:  # Deák Ferenc tér: M1,M2,M3
            r: int = 0
            g: int = 0
            b: int = 0

            if route_status["BKK_5100"]:  # M1 is operational
                r = int(255 * settings.led.dim_ratio)
                g = int(255 * settings.led.dim_ratio)
            if route_status["BKK_5200"]:  # M2 is operational
                r = int(255 * settings.led.dim_ratio)
            if route_status["BKK_5300"]:  # M3 is operational
                b = int(255 * settings.led.dim_ratio)

            DEFAULT_COLORS[led_index] = (r, g, b)
        elif led_index == 22:  # Keleti pályaudvar: M2,M4
            if (
                route_status["BKK_5200"] and route_status["BKK_5400"]
            ):  # Both M2 and M4 are operational
                DEFAULT_COLORS[led_index] = (
                    int(255 * settings.led.dim_ratio),
                    int(128 * settings.led.dim_ratio),
                    0,
                )
            elif route_status["BKK_5200"]:  # Only M2 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_5200"]
            elif route_status["BKK_5400"]:  # Only M4 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_5400"]
            else:
                DEFAULT_COLORS[led_index] = (0, 0, 0)
        elif led_index == 25:  # Örs vezér tere: M2,H8/9
            if (
                route_status["BKK_5200"] and route_status["BKK_H8"]
            ):  # Both M2 and H8 are operational
                DEFAULT_COLORS[led_index] = (
                    int(255 * settings.led.dim_ratio),
                    0,
                    int(48 * settings.led.dim_ratio),
                )
            elif route_status["BKK_5200"]:  # Only M2 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_5200"]
            elif route_status["BKK_H8"]:  # Only H8 is operational
                DEFAULT_COLORS[led_index] = ROUTE_COLORS_DIM["BKK_H8"]
            else:
                DEFAULT_COLORS[led_index] = (0, 0, 0)


def reset_leds_to_default() -> None:
    """Reset the color of all LEDs to the default value."""
    global led_states
    with led_lock:
        led_states = [value for color in DEFAULT_COLORS for value in color]
    # Update the sACN values
    if sender is not None and len(sender.get_active_outputs()) == 1:
        update_sacn()
    logger.debug("All LEDs were reset to their default colors")


def reset_led_to_default(led_index: int, *, fade: bool = True) -> None:
    """Reset the color of a specific LED to the default value.

    :param led_index: The index of the LED
    :param fade: Whether to use fade or not
    """
    if 0 <= led_index < NUM_LEDS:
        if fade:
            with led_lock:
                # Get the current color
                current_color: tuple[int, int, int] = (
                    led_states[led_index * 3],  # Red component
                    led_states[led_index * 3 + 1],  # Green component
                    led_states[led_index * 3 + 2],  # Blue component
                )

            # Calculate fade parameters
            steps: int = int(settings.led.fade_time / 0.005)  # Number of steps
            delay: float = 0.005  # Time to wait between steps

            # Start the fading effect in a separate thread
            threading.Thread(
                target=fade_color,
                args=(
                    led_index,
                    current_color,
                    DEFAULT_COLORS[led_index],
                    steps,
                    delay,
                ),
            ).start()

            logger.trace(  # type: ignore[attr-defined]
                f"LED {led_index} fading from color {current_color!s} "
                f"to default color {DEFAULT_COLORS[led_index]!s}",
            )
        else:
            with led_lock:
                led_states[led_index * 3 : led_index * 3 + 3] = DEFAULT_COLORS[
                    led_index
                ]
            logger.trace(  # type: ignore[attr-defined]
                f"LED {led_index} set color "
                f"to default color {DEFAULT_COLORS[led_index]!s}",
            )
    else:
        logger.error(
            f"Invalid LED index {led_index} when trying to reset the value, "
            f"caller: {inspect.stack()[1].function}",
        )


def set_led_color(led_index: int, color: tuple[int, int, int]) -> None:
    """Reset the color of a specific LED to the supplied RGB tuple.

    :param led_index: The index of the LED
    :param color: A tuple (red, green, blue) representing the RGB values (0-255)
    """
    if 0 <= led_index < NUM_LEDS:
        with led_lock:
            # Get the current color
            current_color: tuple[int, int, int] = (
                led_states[led_index * 3],  # Red component
                led_states[led_index * 3 + 1],  # Green component
                led_states[led_index * 3 + 2],  # Blue component
            )
        # Check if the current color matches one of the ROUTE_COLORS
        if current_color in ROUTE_COLORS.values():
            # Combine the current color with the new color
            color = (
                min(color[0] + current_color[0], 255),
                min(color[1] + current_color[1], 255),
                min(color[2] + current_color[2], 255),
            )

        # Calculate fade parameters
        steps: int = int(settings.led.fade_time / 0.05)  # Number of steps
        delay: float = 0.05  # Time to wait between steps

        # Start the fading effect in a separate thread
        threading.Thread(
            target=fade_color,
            args=(led_index, current_color, color, steps, delay),
        ).start()

        logger.trace(  # type: ignore[attr-defined]
            f"LED {led_index} fading from color {current_color!s} to color {color!s}",
        )
    else:
        logger.error(
            f"Invalid LED index {led_index} when trying to change the value, "
            f"caller: {inspect.stack()[1].function}",
        )


def get_led_color(led_index: int) -> tuple[int, int, int] | None:
    """Retrieve the color of a specific LED as an RGB tuple.

    :param led_index: The index of the LED
    :return: A tuple (red, green, blue) representing the RGB values (0-255)
        or None if the led_index is invalid
    """
    if 0 <= led_index < NUM_LEDS:
        with led_lock:
            return (
                led_states[led_index * 3],
                led_states[led_index * 3 + 1],
                led_states[led_index * 3 + 2],
            )
    else:
        logger.error(
            f"Invalid LED index {led_index} when trying to get the value, "
            f"caller: {inspect.stack()[1].function}",
        )
        return None


def activate_sacn() -> None:
    """Start the sACN sender."""
    global sender
    sender = sacn.sACNsender(
        source_name=f"BudapestMetroDisplay {__version__}",
        cid=tuple(
            uuid.uuid5(
                uuid.UUID("12345678-1234-5678-1234-567812345678"),
                "BudapestMetroDisplay",
            ).bytes,
        ),
        fps=settings.sacn.fps,
    )
    sender.start()

    sender.activate_output(settings.sacn.universe)
    sender[settings.sacn.universe].multicast = settings.sacn.multicast
    if not settings.sacn.multicast:
        sender[settings.sacn.universe].destination = str(settings.sacn.unicast_ip)
    update_sacn()
    logger.info(
        f"sACN settings: "
        f"{'multicast' if sender[settings.sacn.universe].multicast else 'unicast'}, "
        f"destination ip: {sender[settings.sacn.universe].destination}",
    )


def deactivate_sacn() -> None:
    """Stop the sACN sender."""
    if sender is not None:
        sender.stop()


def update_sacn() -> None:
    """Update the internal tuple of the sACN to the latest values.

    Ensure each value is at least 28 when multiplied by esphome.brightness.
    """
    if sender is not None and sender[settings.sacn.universe].dmx_data is not None:
        if settings.esphome.used:
            from BudapestMetroDisplay.esphome import brightness

            # Create a new list with modified values
            modified_led_states: list[int] = [
                (
                    ceil(28 / brightness)
                    if value * brightness < 28 and value != 0
                    else value
                )
                for value in led_states
            ]

            # Convert the modified list to a tuple and assign it to dmx_data
            sender[settings.sacn.universe].dmx_data = tuple(modified_led_states)
        else:
            sender[settings.sacn.universe].dmx_data = tuple(led_states)
